import copy

import networkx as nx
import sympy
from sympy.printing.str import StrPrinter

import pharmpy.symbols as symbols
import pharmpy.unicode as unicode


class MyPrinter(StrPrinter):
    def _print_Add(self, expr):
        args = expr.args
        new = []
        for arg in args:
            new.append(self._print(arg))
        return super()._print_Add(sympy.Add(*args, evaluate=False), order='none')


class Assignment:
    """Representation of variable assignment, similar to :class:`sympy.codegen.Assignment`

    Attributes
    ----------
    symbol : sympy.Symbol
        Symbol of statement
    expression
        Expression of statement
    """

    def __init__(self, symbol, expression):
        """symbol can be either string or sympy symbol
        symbol can be a string and a real symbol will be created
        """
        try:
            symbol.is_Symbol
            self.symbol = symbol
        except AttributeError:
            self.symbol = symbols.symbol(symbol)
        self.expression = sympy.sympify(expression)

    def subs(self, substitutions):
        """Substitute symbols in assignment.

        substitutions - dictionary with old-new pair (can be type str or
                        sympy symbol)
        """
        self.symbol = self.symbol.subs(substitutions, simultaneous=True)
        self.expression = self.expression.subs(substitutions, simultaneous=True)

    @property
    def free_symbols(self):
        symbols = {self.symbol}
        symbols |= self.expression.free_symbols
        return symbols

    @property
    def rhs_symbols(self):
        return self.expression.free_symbols

    def __eq__(self, other):
        return (
            isinstance(other, Assignment)
            and self.symbol == other.symbol
            and self.expression == other.expression
        )

    def __str__(self):
        expression = sympy.pretty(self.expression)
        lines = expression.split('\n')
        definition = f'{sympy.pretty(self.symbol)} := '
        s = ''
        for line in lines:
            if line == lines[-1]:
                s += definition + line + '\n'
            else:
                s += len(definition) * ' ' + line + '\n'
        return s

    def __deepcopy__(self, memo):
        return type(self)(self.symbol, self.expression)

    def __repr__(self):
        return f'{self.symbol} := {self.expression}'

    def _repr_latex_(self):
        sym = self.symbol._repr_latex_()[1:-1]
        expr = self.expression._repr_latex_()[1:-1]
        return f'${sym} := {expr}$'

    def print_custom(self, rvs, trans):
        expr_ordered = self.order_terms(rvs, trans)
        return f'{self.symbol} = {expr_ordered}'

    def order_terms(self, rvs, trans):
        """Order terms such that random variables are placed last. Currently only supports
        additions."""
        if not isinstance(self.expression, sympy.Add) or rvs is None:
            return self.expression
        # FIXME: This should be in MyPrinter
        rvs_names = [rv.name for rv in rvs]

        if trans:
            trans_rvs = {v.name: k.name for k, v in trans.items() if str(k) in rvs_names}

        expr_args = self.expression.args
        terms_iiv_iov, terms_ruv, terms = [], [], []

        for arg in expr_args:
            arg_symbs = [s.name for s in arg.free_symbols]
            rvs_intersect = set(rvs_names).intersection(arg_symbs)

            if trans:
                trans_intersect = set(trans_rvs.keys()).intersection(arg_symbs)
                rvs_intersect.update({trans_rvs[rv] for rv in trans_intersect})

            if rvs_intersect:
                if len(rvs_intersect) == 1:
                    rv_name = list(rvs_intersect)[0]
                    variability_level = rvs[rv_name].level
                    if variability_level == 'RUV':
                        terms_ruv.append(arg)
                        continue
                terms_iiv_iov.append(arg)
            else:
                terms.append(arg)

        if not terms_iiv_iov and not terms_ruv:
            return self.expression

        def arg_len(symb):
            return len([s for s in symb.args])

        terms_iiv_iov.sort(reverse=True, key=arg_len)
        terms_ruv.sort(reverse=True, key=arg_len)
        terms += terms_iiv_iov + terms_ruv

        new_order = sympy.Add(*terms, evaluate=False)

        return MyPrinter().doprint(new_order)


class ODESystem:
    """Base class and placeholder for ODE systems of different forms

    Attributes
    ----------
    solver : str
        Solver to use when numerically solving the ode system
        Currently supports NONMEM ADVANs
    """

    def __init__(self):
        self.solver = None

    @property
    def free_symbols(self):
        return set()

    @property
    def rhs_symbols(self):
        return set()

    def subs(self, substitutions):
        pass

    def __eq__(self, other):
        return isinstance(other, ODESystem)

    def __str__(self):
        return 'ODESystem()'

    def _repr_html(self):
        return str(self)


def _bracket(a):
    """Append a left bracket for an array of lines"""
    if len(a) == 1:
        return '{' + a[0]
    if len(a) == 2:
        a.append('')
    if (len(a) % 2) == 0:
        upper = len(a) // 2 - 1
    else:
        upper = len(a) // 2
    a[0] = '⎧' + a[0]
    for i in range(1, upper):
        a[i] = '⎪' + a[i]
    a[upper] = '⎨' + a[upper]
    for i in range(upper + 1, len(a) - 1):
        a[i] = '⎪' + a[i]
    a[-1] = '⎩' + a[-1]
    return '\n'.join(a) + '\n'


class ExplicitODESystem(ODESystem):
    """System of ODEs described explicitly"""

    def __init__(self, odes, ics):
        self.odes = odes
        self.ics = ics
        super().__init__()

    @property
    def free_symbols(self):
        free = set()
        for ode in self.odes:
            free |= ode.free_symbols
        for key, value in self.ics.items():
            free |= key.free_symbols
            try:  # To allow for regular python classes as values for ics
                free |= value.free_symbols
            except AttributeError:
                pass
        return free

    def subs(self, substitutions):
        d = {
            sympy.Function(str(key))(symbols.symbol('t')): value
            for key, value in substitutions.items()
        }
        self.odes = [ode.subs(d) for ode in self.odes]
        self.ics = {key.subs(d): value.subs(d) for key, value in self.ics.items()}

    @property
    def rhs_symbols(self):
        return self.free_symbols

    def __str__(self):
        a = []
        for ode in self.odes:
            ode_str = sympy.pretty(ode)
            a += ode_str.split('\n')
        for key, value in self.ics.items():
            ics_str = sympy.pretty(sympy.Eq(key, value))
            a += ics_str.split('\n')
        return _bracket(a)

    def __deepcopy__(self, memo):
        newone = type(self)(copy.copy(self.odes), copy.copy(self.ics))
        return newone

    def __eq__(self, other):
        return (
            isinstance(other, ExplicitODESystem)
            and self.odes == other.odes
            and self.ics == other.ics
            and self.solver == other.solver
        )

    def _repr_latex_(self):
        rows = []
        for ode in self.odes:
            ode_repr = ode._repr_latex_()[1:-1]
            rows.append(ode_repr)
        for k, v in self.ics.items():
            ics_eq = sympy.Eq(k, v)
            ics_repr = ics_eq._repr_latex_()[1:-1]
            rows.append(ics_repr)
        return r'\begin{cases} ' + r' \\ '.join(rows) + r' \end{cases}'

    def to_compartmental_system(self):
        funcs = [eq.lhs.args[0] for eq in self.odes]
        cs = CompartmentalSystem()
        for f in funcs:
            cs.add_compartment(f.name)

        for eq in self.odes:
            for comp_func in funcs:
                dep = eq.rhs.as_independent(comp_func, as_Add=True)[1]
                if dep == 0:
                    continue
                terms = sympy.Add.make_args(dep)
                for term in terms:
                    expr = term / comp_func
                    if term.args[0] != -1:
                        from_comp = cs.find_compartment(comp_func.name)
                        to_comp = cs.find_compartment(eq.lhs.args[0].name)
                        cs.add_flow(from_comp, to_comp, expr)

        dose = Bolus("AMT")  # FIXME: not true!
        cs.find_compartment(funcs[0].name).dose = dose
        return cs


class CompartmentalSystem(ODESystem):
    """System of ODEs descibed as a compartmental system"""

    t = symbols.symbol('t')

    def __init__(self):
        self._g = nx.DiGraph()
        super().__init__()

    def subs(self, substitutions):
        for (u, v, rate) in self._g.edges.data('rate'):
            rate_sub = rate.subs(substitutions, simultaneous=True)
            self._g.edges[u, v]['rate'] = rate_sub
        for comp in self._g.nodes:
            comp.subs(substitutions)

    @property
    def free_symbols(self):
        free = {symbols.symbol('t')}
        for (_, _, rate) in self._g.edges.data('rate'):
            free |= rate.free_symbols
        for node in self._g.nodes:
            free |= node.free_symbols
        return free

    @property
    def rhs_symbols(self):
        return self.free_symbols  # This works currently

    def atoms(self, cls):
        atoms = set()
        for (_, _, rate) in self._g.edges.data('rate'):
            atoms |= rate.atoms(cls)
        return atoms

    def __eq__(self, other):
        return (
            isinstance(other, CompartmentalSystem)
            and nx.to_dict_of_dicts(self._g) == nx.to_dict_of_dicts(other._g)
            and self.find_dosing().dose == other.find_dosing().dose
            and self.solver == other.solver
        )

    def __deepcopy__(self, memo):
        newone = type(self)()
        newone._g = copy.deepcopy(self._g, memo)
        return newone

    def add_compartment(self, name):
        comp = Compartment(name, len(self._g) + 1)
        self._g.add_node(comp)
        return comp

    def remove_compartment(self, compartment):
        self._g.remove_node(compartment)

    def add_flow(self, source, destination, rate):
        self._g.add_edge(source, destination, rate=rate)

    def remove_flow(self, source, destination):
        self._g.remove_edge(source, destination)

    def get_flow(self, source, destination):
        try:
            rate = self._g.edges[source, destination]['rate']
        except KeyError:
            rate = None
        return rate

    def get_compartment_outflows(self, compartment):
        """Generate all flows going out of a compartment"""
        flows = []
        for node in self._g.successors(compartment):
            flow = self.get_flow(compartment, node)
            flows.append((node, flow))
        return flows

    def get_compartment_inflows(self, compartment):
        """Generate all flows going in to a compartment"""
        flows = []
        for node in self._g.predecessors(compartment):
            flow = self.get_flow(node, compartment)
            flows.append((node, flow))
        return flows

    def find_compartment(self, name):
        for comp in self._g.nodes:
            if comp.name == name:
                return comp
        else:
            return None

    def n_connected(self, comp):
        """Get the number of compartments connected to comp"""
        out_comps = {c for c, _ in self.get_compartment_outflows(comp)}
        in_comps = {c for c, _ in self.get_compartment_inflows(comp)}
        return len(out_comps | in_comps)

    def find_output(self):
        """Find the output compartment

        An output compartment is defined to be a compartment that does not have any outward
        flow. A model has to have one and only one output compartment.
        """
        zeroout = [node for node, out_degree in self._g.out_degree() if out_degree == 0]
        if len(zeroout) == 1:
            return zeroout[0]
        else:
            raise ValueError('More than one or zero output compartments')

    def find_dosing(self):
        """Find the dosing compartment

        A dosing compartment is a compartment that receives an input dose. Only one dose
        compartment is supported.
        """
        for node in self._g.nodes:
            if node.dose is not None:
                return node
        raise ValueError('No dosing compartment exists')

    def find_central(self):
        """Find the central compartment

        The central compartment is defined to be the compartment that has an outward flow
        to the output compartment. Only one central compartment is supported.
        """
        output = self.find_output()
        central = next(self._g.predecessors(output))
        return central

    def find_peripherals(self):
        central = self.find_central()
        oneout = {node for node, out_degree in self._g.out_degree() if out_degree == 1}
        onein = {node for node, in_degree in self._g.in_degree() if in_degree == 1}
        cout = {comp for comp in oneout if self.get_flow(comp, central) is not None}
        cin = {comp for comp in onein if self.get_flow(central, comp) is not None}
        peripherals = list(cout & cin)
        # Return in deterministic indexed order
        peripherals = sorted(peripherals, key=lambda comp: comp.index)
        return peripherals

    def find_transit_compartments(self, statements):
        """Find all transit compartments

        Transit compartments are a chain of compartments with the same out rate starting from
        the dose compartment. Because one single transit compartment cannot be distinguished
        from one depot compartment such compartment will be defined to be a depot and not
        a transit compartment.
        """
        transits = []
        comp = self.find_dosing()
        if len(self.get_compartment_inflows(comp)) != 0:
            return transits
        outflows = self.get_compartment_outflows(comp)
        if len(outflows) != 1:
            return transits
        transits.append(comp)
        comp, rate = outflows[0]
        rate = statements.full_expression_from_odes(rate)
        while True:
            if len(self.get_compartment_inflows(comp)) != 1:
                break
            outflows = self.get_compartment_outflows(comp)
            if len(outflows) != 1:
                break
            next_comp, next_rate = outflows[0]
            next_rate = statements.full_expression_from_odes(next_rate)
            if rate != next_rate:
                break
            transits.append(comp)
            comp = next_comp
        # Special case of one transit directly into central is not defined as a transit
        # Also not central itself
        central = self.find_central()
        if len(transits) == 1 and (
            self.get_flow(transits[0], central) is not None or transits[0] == central
        ):
            return []
        else:
            return transits

    def find_depot(self, statements):
        """Find the depot compartment

        The depot compartment is defined to be the compartment that only has out flow to the
        central compartment, but no flow from the central compartment.
        """
        transits = self.find_transit_compartments(statements)
        depot = self._find_depot()
        if depot in transits:
            depot = None
        return depot

    def _find_depot(self):
        central = self.find_central()
        depot = None
        for to_central, _ in self.get_compartment_inflows(central):
            outflows = self.get_compartment_outflows(to_central)
            if len(outflows) == 1:
                inflows = self.get_compartment_inflows(to_central)
                for in_comp, _ in inflows:
                    if in_comp == central:
                        break
                else:
                    depot = to_central
                    break
        return depot

    @property
    def compartmental_matrix(self):
        dod = nx.to_dict_of_dicts(self._g)
        size = len(self._g.nodes)
        f = sympy.zeros(size)
        nodes = list(self._g.nodes)
        output = self.find_output()
        # Put output last
        nodes.remove(output)
        nodes.append(output)
        for i in range(0, size):
            from_comp = nodes[i]
            diagsum = 0
            for j in range(0, size):
                to_comp = nodes[j]
                try:
                    rate = dod[from_comp][to_comp]['rate']
                except KeyError:
                    rate = 0
                f[j, i] = rate
                diagsum += f[j, i]
            f[i, i] -= diagsum
        return f

    @property
    def amounts(self):
        amts = [node.amount for node in self._g.nodes]
        return sympy.Matrix(amts)

    @property
    def names(self):
        """A list of the names of all compartments"""
        return [node.name for node in self._g.nodes]

    @property
    def zero_order_inputs(self):
        """A vector of all zero order inputs to each compartment"""
        inputs = []
        for node in self._g.nodes:
            if node.dose is not None and isinstance(node.dose, Infusion):
                if node.dose.rate is not None:
                    expr = node.dose.rate
                    cond = node.dose.amount / node.dose.rate
                else:
                    expr = node.dose.amount / node.dose.duration
                    cond = node.dose.duration
                infusion_func = sympy.Piecewise((expr, self.t < cond), (0, True))
                inputs.append(infusion_func)
            else:
                inputs.append(0)
        return sympy.Matrix(inputs)

    def to_explicit_odes(self, skip_output=False):
        amount_funcs = sympy.Matrix([sympy.Function(amt.name)(self.t) for amt in self.amounts])
        derivatives = sympy.Matrix([sympy.Derivative(fn, self.t) for fn in amount_funcs])
        inputs = self.zero_order_inputs
        a = self.compartmental_matrix @ amount_funcs + inputs
        eqs = [sympy.Eq(lhs, rhs) for lhs, rhs in zip(derivatives, a)]
        ics = {}
        output = self.find_output()
        for node in self._g.nodes:
            if skip_output and node == output:
                continue
            if node.dose is not None and isinstance(node.dose, Bolus):
                if node.lag_time:
                    time = node.lag_time
                else:
                    time = 0
                ics[sympy.Function(node.amount.name)(time)] = node.dose.amount
            else:
                ics[sympy.Function(node.amount.name)(0)] = sympy.Integer(0)
        if skip_output:
            eqs = eqs[:-1]
        return eqs, ics

    def __len__(self):
        """Get the number of compartments including output"""
        return len(self._g.nodes)

    def _repr_html_(self):
        # Use Unicode art for now. There should be ways of drawing networkx
        s = str(self)
        return f'<pre>{s}</pre>'

    def __str__(self):
        output = self.find_output()
        output_box = unicode.Box(output.name)
        central = self.find_central()
        central_box = unicode.Box(central.name)
        depot = self._find_depot()
        current = self.find_dosing()
        if depot:
            comp = depot
        else:
            comp = central
        transits = []
        while current != comp:
            transits.append(current)
            current = self.get_compartment_outflows(current)[0][0]
        periphs = self.find_peripherals()
        nrows = 1 + 2 * len(periphs)
        ncols = 2 * len(transits) + (2 if depot else 0) + 3
        grid = unicode.Grid(nrows, ncols)
        if nrows == 1:
            main_row = 0
        else:
            main_row = 2
        col = 0
        for transit in transits:
            grid.set(main_row, col, unicode.Box(transit.name))
            col += 1
            grid.set(
                main_row, col, unicode.Arrow(str(self.get_compartment_outflows(transit)[0][1]))
            )
            col += 1
        if depot:
            grid.set(main_row, col, unicode.Box(depot.name))
            col += 1
            grid.set(main_row, col, unicode.Arrow(str(self.get_compartment_outflows(depot)[0][1])))
            col += 1
        central_col = col
        grid.set(main_row, col, central_box)
        col += 1
        grid.set(main_row, col, unicode.Arrow(str(self.get_flow(central, output))))
        col += 1
        grid.set(main_row, col, output_box)
        if periphs:
            grid.set(0, central_col, unicode.Box(periphs[0].name))
            grid.set(
                1,
                central_col,
                unicode.DualVerticalArrows(
                    str(self.get_flow(central, periphs[0])), str(self.get_flow(periphs[0], central))
                ),
            )
        if len(periphs) > 1:
            grid.set(4, central_col, unicode.Box(periphs[1].name))
            grid.set(
                3,
                central_col,
                unicode.DualVerticalArrows(
                    str(self.get_flow(periphs[1], central)), str(self.get_flow(central, periphs[1]))
                ),
            )

        dose = self.find_dosing().dose
        return str(dose) + '\n' + str(grid)


class Compartment:
    def __init__(self, name, index, lag_time=0):
        self.name = name
        self.index = index
        self.dose = None
        self.lag_time = lag_time

    @property
    def lag_time(self):
        return self._lag_time

    @lag_time.setter
    def lag_time(self, value):
        self._lag_time = sympy.sympify(value)

    @property
    def free_symbols(self):
        symbs = set()
        if self.dose is not None:
            symbs |= self.dose.free_symbols
        symbs |= self.lag_time.free_symbols
        return symbs

    def subs(self, substitutions):
        if self.dose is not None:
            self.dose.subs(substitutions)
        self.lag_time.subs(substitutions)

    def __eq__(self, other):
        return (
            isinstance(other, Compartment)
            and self.name == other.name
            and self.dose == other.dose
            and self.lag_time == other.lag_time
        )

    def __hash__(self):
        return hash(self.name)

    @property
    def amount(self):
        return symbols.symbol(f'A_{self.name}')


class Bolus:
    def __init__(self, amount):
        self.amount = symbols.symbol(str(amount))

    @property
    def free_symbols(self):
        return {self.amount}

    def subs(self, substitutions):
        self.amount = self.amount.subs(substitutions, simultaneous=True)

    def __deepcopy__(self, memo):
        newone = type(self)(self.amount)
        return newone

    def __eq__(self, other):
        return isinstance(other, Bolus) and self.amount == other.amount

    def __repr__(self):
        return f'Bolus({self.amount})'


class Infusion:
    def __init__(self, amount, rate=None, duration=None):
        if rate is None and duration is None:
            raise ValueError('Need rate or duration for Infusion')
        self.rate = sympy.sympify(rate)
        self.duration = sympy.sympify(duration)
        self.amount = sympy.sympify(amount)

    @property
    def free_symbols(self):
        if self.rate is not None:
            symbs = self.rate.free_symbols
        else:
            symbs = self.duration.free_symbols
        return symbs | self.amount.free_symbols

    def subs(self, substitutions):
        self.amount = self.amount.subs(substitutions, simultaneous=True)
        if self.rate is not None:
            self.rate = self.rate.subs(substitutions, simultaneous=True)
        else:
            self.duration = self.duration.subs(substitutions, simultaneous=True)

    def __deepcopy__(self, memo):
        new = type(self)(self.amount, rate=self.rate, duration=self.duration)
        return new

    def __eq__(self, other):
        return (
            isinstance(other, Infusion)
            and self.rate == other.rate
            and self.duration == other.duration
            and self.amount == other.amount
        )

    def __repr__(self):
        if self.rate is not None:
            arg = f'rate={self.rate}'
        else:
            arg = f'duration={self.duration}'
        return f'Infusion({self.amount}, {arg})'


class ModelStatements(list):
    """A list of sympy statements describing the model"""

    @property
    def free_symbols(self):
        """Get a set of all free symbols"""
        symbols = set()
        for assignment in self:
            symbols |= assignment.free_symbols
        return symbols

    def subs(self, substitutions):
        """Substitute symbols in all statements.

        substitutions - dictionary with old-new pair (can be type str or
                        sympy symbol)
        """
        for assignment in self:
            assignment.subs(substitutions)

    def find_assignment(self, variable, is_symbol=True, last=True):
        """Returns full last statement or all assignments that contains the symbol or
        variable of interest of an assignment"""
        assignments = []
        for statement in self:
            if isinstance(statement, Assignment):
                if is_symbol and str(statement.symbol) == variable:
                    assignments.append(statement)
                elif not is_symbol and variable in [s.name for s in statement.free_symbols]:
                    assignments.append(statement)

        if last:
            try:
                return assignments[-1]
            except IndexError:
                return None
        else:
            return assignments

    def extract_params_from_symb(self, symbol_name, pset):
        terms = {symb.name for symb in self.find_assignment(symbol_name).free_symbols}
        theta_name = terms.intersection(pset.names).pop()
        return pset[theta_name]

    def reassign(self, symbol, expression):
        """Reassign symbol to expression"""
        last = True
        for i, stat in zip(range(len(self) - 1, -1, -1), reversed(self)):
            if isinstance(stat, Assignment) and stat.symbol == symbol:
                if last:
                    stat.expression = expression
                    last = False
                else:
                    del self[i]

    def _create_dependency_graph(self):
        """Create a graph of dependencies between assignments
        Does not support dependencies of ODESystem
        """
        graph = nx.DiGraph()
        for i in range(len(self) - 1, -1, -1):
            rhs = self[i].rhs_symbols
            for s in rhs:
                for j in range(i - 1, -1, -1):
                    if isinstance(self[j], Assignment) and self[j].symbol == s:
                        graph.add_edge(i, j)
                        break
        return graph

    def remove_symbol_definitions(self, symbols, statement):
        """Remove symbols and dependencies not used elsewhere. statement
        is the statement from which the symbol was removed
        """
        graph = self._create_dependency_graph()
        removed_ind = self.index(statement)
        # Statements defining symbols and dependencies
        candidates = set()
        for s in symbols:
            for i in range(removed_ind - 1, -1, -1):
                stat = self[i]
                if isinstance(stat, Assignment) and stat.symbol == s:
                    candidates.add(i)
                    break
        for i in candidates.copy():
            if i in graph:
                candidates |= set(nx.dfs_preorder_nodes(graph, i))
        # All statements needed for removed_ind
        if removed_ind in graph:
            keep = {down for _, down in nx.dfs_edges(graph, removed_ind)}
        else:
            keep = set()
        candidates -= keep
        # Other dependencies after removed_ind
        additional = {down for up, down in graph.edges if up > removed_ind and down in candidates}
        for add in additional.copy():
            if add in graph:
                additional |= set(nx.dfs_preorder_nodes(graph, add))
        remove = candidates - additional
        for i in reversed(sorted(remove)):
            del self[i]

    @property
    def ode_system(self):
        """Returns the ODE system of the model or None if the model doesn't have an ODE system"""
        for s in self:
            if isinstance(s, ODESystem):
                return s
        return None

    def before_ode(self):
        sset = ModelStatements()
        for s in self:
            if isinstance(s, ODESystem):
                break
            sset.append(s)

        return sset

    def _ode_index(self):
        for i, s in enumerate(self):
            if isinstance(s, ODESystem):
                return i
        return None

    def full_expression_from_odes(self, expression):
        """Expand an expression into its full definition

        Before ODE system
        """
        i = self._ode_index()
        if i is None:
            i = len(self)
        for j in range(i - 1, -1, -1):
            expression = expression.subs({self[j].symbol: self[j].expression})
        return expression

    def full_expression_after_odes(self, expression):
        """Expand an expression into its full definition

        After ODE system
        """
        i = self._ode_index()
        if i is None:
            i = -1
        for j in range(len(self) - 1, i, -1):
            expression = expression.subs({self[j].symbol: self[j].expression})
        return expression

    def add_before_odes(self, statement):
        """Add a statement just before the ODE system or at the end of the model"""
        for i, s in enumerate(self):
            if isinstance(s, ODESystem):
                break
        else:
            i += 1
        self.insert(i, statement)

    def copy(self):
        return copy.deepcopy(self)

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        else:
            for i in range(len(self)):
                if self[i] != other[i]:
                    return False
        return True

    def __repr__(self):
        s = ''
        for statement in self:
            s += repr(statement) + '\n'
        return s

    def __str__(self):
        s = ''
        for statement in self:
            s += str(statement)
        return s

    def _repr_html_(self):
        html = r'\begin{align*}'
        for statement in self:
            if hasattr(statement, '_repr_html_'):
                html += '\\end{align*}'
                s = statement._repr_html_()
                html += s + '\\begin{align*}'
            else:
                s = f'${statement._repr_latex_()}$'
                s = s.replace(':=', '&:=')
                s = s.replace('$', '')
                s = s + r'\\'
                html += s
        return html + '\\end{align*}'



import pharmpy.tools


class Modelfit(pharmpy.tools.Tool):
    def __init__(self, models, **kwargs):
        self.models = models
        super().__init__(**kwargs)

    def run(self):
        workflow = self.workflow_creator(self.models)
        fit_models = self.dispatcher.run(workflow, self.database)
        for i in range(len(fit_models)):
            self.models[i].modelfit_results = fit_models[i].modelfit_results
        # res = self.models[0].modelfit_results
        # res.to_json(path=self.rundir.path / 'results.json')
        # res.to_csv(path=self.rundir.path / 'results.csv')



from io import StringIO

import pandas as pd
import pytest

from pharmpy import Model
from pharmpy.results import read_results
from pharmpy.tools.qa.results import calculate_results


def test_add_etas(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    add_etas = Model(testdata / 'nonmem' / 'qa' / 'add_etas_linbase.mod')
    res = calculate_results(orig, base, add_etas_model=add_etas, etas_added_to=['CL', 'V'])
    correct = """added,new_sd,orig_sd
ETA(1),True,0.338974,0.333246
ETA(2),True,0.449430,0.448917
CL,False,0.010001,NaN
V,False,0.010000,NaN
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.add_etas_parameters, correct, atol=1e-6)

    assert res.dofv['dofv']['parameter_variability', 'add_etas'] == pytest.approx(
        730.89472681373070 - 730.84697789365532
    )
    assert res.dofv['df']['parameter_variability', 'add_etas'] == 2


def test_fullblock(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    fb = Model(testdata / 'nonmem' / 'qa' / 'fullblock.mod')
    res = calculate_results(orig, base, fullblock_model=fb)
    correct = """,new,old
"OMEGA(1,1)",0.486600,0.333246
"OMEGA(2,1)",0.846728,NaN
"OMEGA(2,2)",0.423262,0.448917
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.fullblock_parameters, correct)

    assert res.dofv['dofv']['parameter_variability', 'fullblock'] == pytest.approx(
        730.89472681373070 - 706.36113798726512
    )
    assert res.dofv['df']['parameter_variability', 'fullblock'] == 1

    res = calculate_results(orig, base, fullblock_model=None)
    assert res.fullblock_parameters is None


def test_boxcox(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    bc = Model(testdata / 'nonmem' / 'qa' / 'boxcox.mod')
    res = calculate_results(orig, base, boxcox_model=bc)
    correct = """lambda,new_sd,old_sd
ETA(1),-1.581460,0.296257,0.333246
ETA(2),0.645817,0.429369,0.448917
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.boxcox_parameters, correct)

    assert res.dofv['dofv']['parameter_variability', 'boxcox'] == pytest.approx(
        730.89472681373070 - 721.78812733817688
    )
    assert res.dofv['df']['parameter_variability', 'boxcox'] == 2

    res = calculate_results(orig, base, boxcox_model=None)
    assert res.boxcox_parameters is None


def test_tdist(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    td = Model(testdata / 'nonmem' / 'qa' / 'tdist.mod')
    res = calculate_results(orig, base, tdist_model=td)
    correct = """df,new_sd,old_sd
ETA(1),3.77,0.344951,0.333246
ETA(2),3.77,0.400863,0.448917
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.tdist_parameters, correct)

    assert res.dofv['dofv']['parameter_variability', 'tdist'] == pytest.approx(
        730.89472681373070 - 729.45800311609150
    )
    assert res.dofv['df']['parameter_variability', 'tdist'] == 2

    res = calculate_results(orig, base, tdist_model=None)


def test_iov(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    iov = Model(testdata / 'nonmem' / 'qa' / 'iov.mod')
    res = calculate_results(orig, base, iov_model=iov)
    correct = """new_iiv_sd,orig_iiv_sd,iov_sd
ETA(1),0.259560,0.333246,0.555607
ETA(2),0.071481,0.448917,0.400451
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.iov_parameters, correct)

    assert res.dofv['dofv']['parameter_variability', 'iov'] == pytest.approx(42.314986)
    assert res.dofv['df']['parameter_variability', 'iov'] == 2


def test_scm(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    scm_res = read_results(testdata / 'nonmem' / 'qa' / 'scm_results.json')
    res = calculate_results(orig, base, scm_results=scm_res)
    correct = """,,dofv,coeff
ETA(1),APGR,2.48792,-0.033334
ETA(1),WGT,0.48218,0.052342
ETA(2),APGR,0.59036,0.008371
ETA(2),WGT,0.00887,-0.003273
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0, 1])
    correct.index.set_names(['parameter', 'covariate'], inplace=True)
    pd.testing.assert_frame_equal(res.covariate_effects, correct, atol=1e-6)
    assert res.dofv['dofv']['covariates', 'ET1APGR-2'] == pytest.approx(2.48792)
    assert res.dofv['df']['covariates', 'ET1APGR-2'] == 1



import numpy as np
import pytest

from pharmpy import Model


def test_individual_parameter_statistics(testdata):
    model = Model(testdata / 'nonmem' / 'secondary_parameters' / 'pheno.mod')
    rng = np.random.default_rng(103)
    stats = model.modelfit_results.individual_parameter_statistics('CL/V', seed=rng)

    assert stats['mean'][0] == pytest.approx(0.004700589484324183)
    assert stats['variance'][0] == pytest.approx(8.086653508585209e-06)
    assert stats['stderr'][0] == pytest.approx(0.0035089729730046304, abs=1e-6)

    model = Model(testdata / 'nonmem' / 'secondary_parameters' / 'run1.mod')
    rng = np.random.default_rng(5678)
    stats = model.modelfit_results.individual_parameter_statistics('CL/V', seed=rng)
    assert stats['mean'][0] == pytest.approx(0.0049100899539843)
    assert stats['variance'][0] == pytest.approx(7.391076132098555e-07)
    assert stats['stderr'][0] == pytest.approx(0.0009425952783595735, abs=1e-6)

    covmodel = Model(testdata / 'nonmem' / 'secondary_parameters' / 'run2.mod')
    rng = np.random.default_rng(8976)
    stats = covmodel.modelfit_results.individual_parameter_statistics('K = CL/V', seed=rng)
    assert stats['mean']['K', 'median'] == pytest.approx(0.004525842355027405)
    assert stats['variance']['K', 'median'] == pytest.approx(2.9540381716908423e-06)
    assert stats['stderr']['K', 'median'] == pytest.approx(0.001804371451706786, abs=1e-6)
    assert stats['mean']['K', 'p5'] == pytest.approx(0.0033049497924269385)
    assert stats['variance']['K', 'p5'] == pytest.approx(1.5730213328583985e-06)
    assert stats['stderr']['K', 'p5'] == pytest.approx(0.0013102577338191103, abs=1e-6)
    assert stats['mean']['K', 'p95'] == pytest.approx(0.014625434302866478)
    assert stats['variance']['K', 'p95'] == pytest.approx(3.090546438695198e-05)
    assert stats['stderr']['K', 'p95'] == pytest.approx(0.0069971678916412716, abs=1e-6)


def test_pk_parameters(testdata):
    model = Model(testdata / 'nonmem' / 'models' / 'mox1.mod')
    rng = np.random.default_rng(103)
    df = model.modelfit_results.pk_parameters(seed=rng)
    assert df['mean'].loc['t_max', 'median'] == pytest.approx(1.5999856886869577)
    assert df['variance'].loc['t_max', 'median'] == pytest.approx(0.29728565293669557)
    assert df['stderr'].loc['t_max', 'median'] == pytest.approx(0.589128711884761)
    assert df['mean'].loc['C_max_dose', 'median'] == pytest.approx(0.6306393134647171)
    assert df['variance'].loc['C_max_dose', 'median'] == pytest.approx(0.012194951111832641)
    assert df['stderr'].loc['C_max_dose', 'median'] == pytest.approx(0.11328030491204867)



import numpy as np
import pandas as pd

from .mfl import ModelFeatures


def exhaustive(base_model, mfl, run_func, rank_func):
    features = ModelFeatures(mfl)
    torun = []
    combinations = list(features.all_combinations())
    df = pd.DataFrame(
        index=pd.RangeIndex(stop=len(combinations)),
        columns=['features', 'dofv', 'rank'],
    )
    funcs = features.all_funcs()
    for i, combo in enumerate(combinations):
        model = base_model.copy()
        model.name = f'candidate{i}'
        for feat in combo:
            funcs[feat](model)
        df.loc[i]['features'] = tuple(combo)
        torun.append(model)
    run_func(torun)
    for i, model in enumerate(torun):
        df.loc[i]['dofv'] = base_model.modelfit_results.ofv - model.modelfit_results.ofv
    ranks = rank_func(base_model, torun)
    for i, ranked_model in enumerate(ranks):
        idx = torun.index(ranked_model)
        df.loc[idx]['rank'] = i + 1
    df = df.astype({'rank': 'Int64'})
    return df


def stepwise(base_model, mfl, run_func, rank_func):
    features = ModelFeatures(mfl)
    remaining = features.all_funcs()
    start_model = base_model
    current_features = []
    features_col = []
    dofv_col = []
    step = 1
    while True:
        torun = []
        for feat, func in remaining.items():
            model = start_model.copy()
            model.name = f'step_{step}_{feat.replace("(", "_").replace(")", "")}'
            func(model)
            torun.append(model)
            features_col.append(tuple(current_features + [feat]))
        print(f'Running step {step}')
        run_func(torun)
        for model in torun:
            dofv = start_model.modelfit_results.ofv - model.modelfit_results.ofv
            dofv_col.append(dofv)
        ranks = rank_func(start_model, torun)
        if not ranks:
            break
        start_model = ranks[0]
        start_model.update_inits()
        idx = torun.index(start_model)
        selected_feature = list(remaining.keys())[idx]
        current_features.append(selected_feature)
        remaining = features.next_funcs(current_features)
        step += 1
    df = pd.DataFrame({'features': features_col, 'dofv': dofv_col, 'rank': np.nan})
    best_features = tuple(current_features)
    best_df_index = features_col.index(best_features)
    df.at[best_df_index, 'rank'] = 1
    df = df.astype({'rank': 'Int64'})
    return df



import shutil
import tempfile
from pathlib import Path

from pharmpy.utils import TemporaryDirectoryChanger

from ..dispatcher import ExecutionDispatcher
from .local_run import run


class LocalDispatcher(ExecutionDispatcher):
    def run(self, workflow, database):
        with tempfile.TemporaryDirectory() as tempdirname:
            temppath = Path(tempdirname)
            for source, dest in workflow.infiles:
                dest_path = temppath / dest
                dest_path.mkdir(parents=True, exist_ok=True)
                shutil.copy2(source, dest_path)
            with TemporaryDirectoryChanger(temppath):
                results = run(workflow.as_dict())
            for model, files in zip(workflow.models, workflow.outfiles):
                for f in files:
                    database.store_local_file(model, temppath / f)
            return results



"""
:meta private:
"""

import warnings

from pharmpy.random_variables import RandomVariables
from pharmpy.symbols import symbol as S


def remove_iov(model):
    """
    Removes all IOV omegas.

    Parameters
    ----------
    model : Model
        Pharmpy model to remove IOV from.
    """
    rvs, sset = model.random_variables, model.statements
    etas = _get_etas(rvs)

    if not etas:
        warnings.warn('No IOVs present')
        return model

    for eta in etas:
        statement = sset.find_assignment(eta.name, is_symbol=False)
        statement.expression = statement.expression.subs(S(eta.name), 0)
        del rvs[eta]

    model.random_variables = rvs
    model.statements = sset

    model.modelfit_results = None

    return model


def _get_etas(rvs):
    etas = [eta for eta in rvs if eta.level == 'IOV']

    return RandomVariables(etas)



"""
:meta private:
"""

import numpy as np

from pharmpy import math
from pharmpy.modeling.help_functions import _get_etas
from pharmpy.parameter import Parameter
from pharmpy.random_variables import RandomVariables


def create_rv_block(model, list_of_rvs=None):
    """
    Creates a full or partial block structure of etas. The etas must be IIVs and cannot
    be fixed. Initial estimates for covariance between the etas is dependent on whether
    the model has results from a previous results. In that case, the correlation will
    be calculated from individual estimates, otherwise correlation will be set to 10%.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    list_of_rvs : list
        List of etas to create a block structure from. If None, all etas that are IIVs and
        non-fixed will be used (full block). None is default.
    """
    rvs = model.random_variables
    if list_of_rvs is None:
        iiv_rvs = model.random_variables.iiv
        nonfix = RandomVariables()
        for rv in iiv_rvs:
            for name in iiv_rvs.parameter_names:
                if model.parameters[name].fix:
                    break
            else:
                nonfix.append(rv)
        list_of_rvs = nonfix.names
    else:
        for name in list_of_rvs:
            if name in rvs and rvs[name].level == 'IOV':
                raise ValueError(
                    f'{name} describes IOV: Joining IOV random variables is currently not supported'
                )
    if len(list_of_rvs) == 1:
        raise ValueError('At least two random variables are needed')

    sset = model.statements
    paramnames = []
    for rv in list_of_rvs:
        statements = sset.find_assignment(rv, is_symbol=False, last=False)
        parameter_names = '_'.join([s.symbol.name for s in statements])
        paramnames.append(parameter_names)

    cov_to_params = rvs.join(list_of_rvs, name_template='IIV_{}_IIV_{}', param_names=paramnames)

    pset = model.parameters
    for cov_name, param_names in cov_to_params.items():
        parent_params = (pset[param_names[0]], pset[param_names[1]])
        covariance_init = _choose_param_init(model, rvs, parent_params)
        param_new = Parameter(cov_name, covariance_init)
        pset.append(param_new)

    return model


def split_rv_block(model, list_of_rvs=None):
    """
    Splits a block structure given a list of etas to separate.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    list_of_rvs : str, list
        Name/names of etas to split from block structure. If None, all etas that are IIVs and
        non-fixed will become single. None is default.
    """
    rvs = model.random_variables
    list_of_rvs = _get_etas(model, list_of_rvs)

    parameters_before = rvs.parameter_names
    rvs.unjoin(list_of_rvs)
    parameters_after = rvs.parameter_names

    removed_parameters = set(parameters_before) - set(parameters_after)
    pset = model.parameters
    for param in removed_parameters:
        del pset[param]

    return model


def _choose_param_init(model, rvs, params):
    res = model.modelfit_results
    rvs_names = [rv.name for rv in rvs]

    etas = []
    for i in range(len(rvs)):
        elem = rvs.covariance_matrix.row(i).col(i)[0]
        if str(elem) in [p.name for p in params]:
            etas.append(rvs_names[i])

    sd = np.array([np.sqrt(params[0].init), np.sqrt(params[1].init)])
    init_default = round(0.1 * sd[0] * sd[1], 7)

    if res is not None:
        try:
            ie = res.individual_estimates
            if not all(eta in ie.columns for eta in etas):
                return init_default
        except KeyError:
            return init_default
        eta_corr = ie[etas].corr()
        cov = math.corr2cov(eta_corr.to_numpy(), sd)
        cov[cov == 0] = 0.0001
        cov = math.nearest_posdef(cov)
        return round(cov[1][0], 7)
    else:
        return init_default



from pharmpy.results import Results


class ResmodResults(Results):
    """Resmod results class"""

    def __init__(self):
        pass


def calculate_results():
    """Calculate resmod results"""
    res = ResmodResults()
    return res


def psn_resmod_results():
    res = calculate_results()
    return res



import pharmpy.tools
import pharmpy.tools.modelfit as modelfit
from pharmpy.data.iterators import Resample


class Bootstrap(pharmpy.tools.Tool):
    def __init__(self, model):
        self.model = model
        super().__init__()

    def run(self):
        resample = Resample(self.model, self.model.dataset.pharmpy.id_label, name_pattern='bs_{}')
        models = []
        for (remod, groups) in resample:
            remod.write(path=self.rundir.models_path)  # FIXME: Automatic save to models?
            models.append(remod)
        modelfit_run = modelfit.Modelfit(models, path=self.rundir.path)
        modelfit_run.run()

        # run(self.models, self.rundir.path)
        # res = self.models[0].modelfit_results
        # res.to_json(path=self.rundir.path / 'results.json')
        # res.to_csv(path=self.rundir.path / 'results.csv')



"""
:meta private:
"""

from pharmpy.modeling import has_proportional_error
from pharmpy.modeling.help_functions import _format_input_list
from pharmpy.parameter import Parameter
from pharmpy.symbols import symbol as S


def power_on_ruv(model, list_of_eps=None):
    """
    Applies a power effect to provided epsilons. Initial estimates for new thetas are 1 if the error
    model is proportional, otherwise they are 0.1.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    list_of_eps : str, list
        Name/names of epsilons to apply power effect. If None, all epsilons will be used.
        None is default.
    """
    list_of_eps = _format_input_list(list_of_eps)
    eps = model.random_variables.epsilons
    if list_of_eps is not None:
        eps = eps[list_of_eps]
    pset, sset = model.parameters, model.statements

    if has_proportional_error(model):
        theta_init = 1
    else:
        theta_init = 0.1

    for i, e in enumerate(eps):
        theta_name = str(model.create_symbol(stem='power', force_numbering=True))
        theta = Parameter(theta_name, theta_init)
        pset.append(theta)
        sset.subs({e.name: model.individual_prediction_symbol ** S(theta.name) * e.symbol})

    model.parameters = pset
    model.statements = sset

    return model



"""
:meta private:
"""


class ModelfitResultsException(Exception):
    pass


def update_inits(model, force_individual_estimates=False):
    """
    Updates initial estimates from previous output. Can be forced if no initial
    individual estimates have been read.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    force_individual_estimates : bool
        Whether update of initial individual estimates should be forced.
    """
    try:
        res = model.modelfit_results
    except AttributeError:
        raise ModelfitResultsException('No modelfit results available')

    model.parameters = res.parameter_estimates

    if model.initial_individual_estimates is not None or force_individual_estimates:
        model.initial_individual_estimates = res.individual_estimates

    return model



"""Factory for creating a pharmpy model from a "native" object representation.

Definitions
===========
"""
import pharmpy.source as source
from pharmpy.plugins.utils import detect_model


def Model(obj, **kwargs):
    """Factory for creating a :class:`pharmpy.model` object from an object representing the model
    (i.e. path).

    .. _path-like object: https://docs.python.org/3/glossary.html#term-path-like-object

    Parameters
    ----------
    obj
        Currently a `path-like object`_ pointing to the model file.

    Returns
    -------
    - Generic :class:`~pharmpy.generic.Model` if path is None, otherwise appropriate implementation
      is invoked (e.g. NONMEM7 :class:`~pharmpy.api_nonmem.model.Model`).
    """
    src = source.Source(obj)
    model_class = detect_model(src)
    model = model_class(src, **kwargs)
    return model



from pharmpy.modeling.block_rvs import create_rv_block, split_rv_block
from pharmpy.modeling.common import (
    convert_model,
    copy_model,
    fix_parameters,
    fix_parameters_to,
    read_model,
    read_model_from_string,
    set_initial_estimates,
    set_name,
    unfix_parameters,
    unfix_parameters_to,
    update_source,
    write_model,
)
from pharmpy.modeling.covariate_effect import add_covariate_effect
from pharmpy.modeling.error import (
    additive_error,
    combined_error,
    has_additive_error,
    has_combined_error,
    has_proportional_error,
    proportional_error,
    remove_error,
    set_dtbs_error,
    set_weighted_error_model,
    theta_as_stdev,
)
from pharmpy.modeling.eta_additions import add_iiv, add_iov
from pharmpy.modeling.eta_transformations import boxcox, john_draper, tdist
from pharmpy.modeling.evaluation import evaluate_expression
from pharmpy.modeling.iiv_on_ruv import iiv_on_ruv
from pharmpy.modeling.odes import (
    add_lag_time,
    add_parameter,
    add_peripheral_compartment,
    bolus_absorption,
    explicit_odes,
    first_order_absorption,
    first_order_elimination,
    michaelis_menten_elimination,
    mixed_mm_fo_elimination,
    remove_lag_time,
    remove_peripheral_compartment,
    seq_zo_fo_absorption,
    set_ode_solver,
    set_peripheral_compartments,
    set_transit_compartments,
    zero_order_absorption,
    zero_order_elimination,
)
from pharmpy.modeling.power_on_ruv import power_on_ruv
from pharmpy.modeling.remove_iiv import remove_iiv
from pharmpy.modeling.remove_iov import remove_iov
from pharmpy.modeling.run import create_results, fit, read_results
from pharmpy.modeling.update_inits import update_inits

__all__ = [
    'add_parameter',
    'zero_order_absorption',
    'first_order_absorption',
    'bolus_absorption',
    'seq_zo_fo_absorption',
    'add_covariate_effect',
    'add_iiv',
    'add_lag_time',
    'boxcox',
    'create_rv_block',
    'explicit_odes',
    'fix_parameters',
    'iiv_on_ruv',
    'john_draper',
    'remove_lag_time',
    'tdist',
    'unfix_parameters',
    'update_source',
    'read_model',
    'read_model_from_string',
    'write_model',
    'remove_iiv',
    'remove_iov',
    'set_transit_compartments',
    'michaelis_menten_elimination',
    'zero_order_elimination',
    'mixed_mm_fo_elimination',
    'first_order_elimination',
    'additive_error',
    'proportional_error',
    'combined_error',
    'remove_error',
    'add_peripheral_compartment',
    'remove_peripheral_compartment',
    'update_inits',
    'power_on_ruv',
    'fit',
    'set_ode_solver',
    'add_iov',
    'set_initial_estimates',
    'copy_model',
    'set_name',
    'has_proportional_error',
    'has_additive_error',
    'has_combined_error',
    'split_rv_block',
    'fix_parameters_to',
    'unfix_parameters_to',
    'create_results',
    'set_peripheral_compartments',
    'read_results',
    'evaluate_expression',
    'convert_model',
    'theta_as_stdev',
    'set_weighted_error_model',
    'set_dtbs_error',
]



from .database import ModelDatabase
from .databases import LocalDirectoryDatabase, LocalModelDirectoryDatabase
from .dispatcher import ExecutionDispatcher
from .dispatchers import LocalDispatcher

default_dispatcher = LocalDispatcher()
default_database = LocalModelDirectoryDatabase()

__all__ = [
    'LocalDirectoryDatabase',
    'LocalModelDirectoryDatabase',
    'ExecutionDispatcher',
    'ModelDatabase',
]



import pandas as pd

from pharmpy import Model
from pharmpy.results import Results
from pharmpy.tools.simfit.results import psn_simfit_results


class SimevalResults(Results):
    """Simeval results class"""

    def __init__(self, sampled_iofv=None, iofv_summary=None, individual_predictions_plot=None):
        self.sampled_iofv = sampled_iofv
        self.iofv_summary = iofv_summary
        self.individual_predictions_plot = individual_predictions_plot


def calculate_results(original_model, simfit_results):
    """Calculate simeval results"""
    sampled_iofv = pd.DataFrame()
    for i, res in enumerate(simfit_results.modelfit_results):
        sampled_iofv[f'{i + 1}'] = res.individual_ofv
    origiofv = original_model.modelfit_results.individual_ofv
    iofv_summary = pd.DataFrame(
        {
            'original': origiofv,
            'sampled_median': sampled_iofv.T.median(),
            'sampled_stdev': sampled_iofv.T.std(),
        }
    )
    iofv_summary['residual'] = (
        abs(iofv_summary['original'] - iofv_summary['sampled_median'])
        / iofv_summary['sampled_stdev']
    )
    iofv_summary['residual_outlier'] = iofv_summary['residual'] >= 3

    ids = iofv_summary.index[iofv_summary['residual_outlier']].tolist()
    id_plot = None
    if ids:
        try:
            id_plot = original_model.modelfit_results.plot_individual_predictions(
                predictions=['CIPREDI', 'PRED'], individuals=ids
            )
        except Exception:
            pass

    res = SimevalResults(
        sampled_iofv=sampled_iofv,
        iofv_summary=iofv_summary,
        individual_predictions_plot=id_plot,
    )
    return res


def psn_simeval_results(path):
    simfit_paths = (path / 'm1').glob('sim-*.mod')
    simfit_results = psn_simfit_results(simfit_paths)
    original = Model(path / 'm1' / 'original.mod')
    res = calculate_results(original, simfit_results)

    # Add CWRES outliers as 2 in data_flag
    # Reading PsN results for now
    # df = pd.read_csv(path / 'summary_cwres.csv')
    # outliers = df['OUTLIER'].fillna(0).astype(int)
    # outliers.replace({1.0: 2}, inplace=True)
    # outliers = outliers + res.data_flag
    # outliers.replace({3: 1}, inplace=True)
    # res.data_flag = outliers
    return res



import os
from pathlib import Path


class TemporaryDirectoryChanger:
    def __init__(self, path):
        self.path = path
        self.old_path = Path.cwd()

    def __enter__(self):
        os.chdir(self.path)
        return self

    def __exit__(self, *args):
        os.chdir(self.old_path)



import warnings
from functools import partial

import numpy as np
import pandas as pd

from pharmpy.math import is_posdef, nearest_posdef, sample_truncated_joint_normal


def sample_from_function(model, samplingfn, parameters=None, force_posdef_samples=None, n=1):
    """Sample parameter vectors using a general function

    The sampling function will be given three arguments:

    - lower - lower bounds of parameters
    - upper - upper bounds of parameters
    - n - number of samples
    """
    if parameters is None:
        parameters = model.parameters.names

    parameter_summary = model.parameters.to_dataframe().loc[parameters]
    parameter_summary = parameter_summary[~parameter_summary['fix']]
    lower = parameter_summary.lower.astype('float64').to_numpy()
    upper = parameter_summary.upper.astype('float64').to_numpy()

    # reject non-posdef
    kept_samples = pd.DataFrame()
    remaining = n

    if force_posdef_samples == 0:
        force_posdef = True
    else:
        force_posdef = False

    i = 0
    while remaining > 0:
        samples = samplingfn(lower, upper, n=remaining)
        df = pd.DataFrame(samples, columns=parameters)
        if not force_posdef:
            selected = df[df.apply(model.random_variables.validate_parameters, axis=1)]
        else:
            rvs = model.random_variables
            selected = df.transform(lambda row: rvs.nearest_valid_parameters(row), axis=1)
        kept_samples = pd.concat((kept_samples, selected))
        remaining = n - len(kept_samples)
        i += 1
        if not force_posdef and force_posdef_samples is not None and i >= force_posdef_samples:
            force_posdef = True

    return kept_samples.reset_index(drop=True)


def sample_uniformly(
    model, fraction=0.1, parameters=None, force_posdef_samples=None, n=1, seed=None
):
    """Sample parameter vectors using uniform sampling

    Each parameter value will be randomly sampled from a uniform distriution
    with lower bound estimate - estimate * fraction and upper bound
    estimate + estimate * fraction
    """

    if seed is None or isinstance(seed, int):
        seed = np.random.default_rng(seed)

    def fn(lower, upper, n):
        samples = np.empty((n, len(lower)))
        for i, (a, b) in enumerate(zip(lower, upper)):
            samples[i, :] = seed.uniform(a, b, n)
        return samples

    samples = sample_from_function(
        model, fn, parameters=parameters, force_posdef_samples=force_posdef_samples, n=n
    )
    return samples


def sample_from_covariance_matrix(
    model,
    modelfit_results=None,
    parameters=None,
    force_posdef_samples=None,
    force_posdef_covmatrix=False,
    n=1,
    seed=None,
):
    """Sample parameter vectors using the covariance matrix

    If modelfit_results is not provided the results from the model will be used

    Parameters
    ----------
    parameters
        Use to only sample a subset of the parameters. None means all
    force_posdef_samples
        Set to how many iterations to do before forcing all samples to be positive definite. None is
        default and means never and 0 means always

    Returns
    -------
    A dataframe with one sample per row
    """
    if modelfit_results is None:
        modelfit_results = model.modelfit_results

    if parameters is None:
        parameters = list(modelfit_results.parameter_estimates.index)

    if seed is None or isinstance(seed, int):
        seed = np.random.default_rng(seed)

    pe = modelfit_results.parameter_estimates[parameters]
    index = pe.index
    mu = pe.to_numpy()
    sigma = modelfit_results.covariance_matrix[parameters].loc[parameters].to_numpy()
    if not is_posdef(sigma):
        if force_posdef_covmatrix:
            old_sigma = sigma
            sigma = nearest_posdef(sigma)
            delta_frobenius = np.linalg.norm(old_sigma) - np.linalg.norm(sigma)
            delta_max = np.abs(old_sigma).max() - np.abs(sigma).max()
            warnings.warn(
                f'Covariance matrix was forced to become positive definite.\n'
                f'    Difference in the frobenius norm: {delta_frobenius:.3e}\n'
                f'    Difference in the max norm: {delta_max:.3e}\n'
            )
        else:
            raise ValueError("Uncertainty covariance matrix not positive-definite")

    fn = partial(sample_truncated_joint_normal, mu, sigma, seed=seed)
    samples = sample_from_function(
        model, fn, parameters=index, force_posdef_samples=force_posdef_samples, n=n
    )
    return samples


def sample_individual_estimates(model, parameters=None, samples_per_id=100, seed=None):
    """Sample individual estimates given their covariance.

    Parameters
    ----------
    parameters
        A list of a subset of parameters to sample. Default is None, which means all.

    Returns
    -------
    Pool of samples in a DataFrame
    """
    if seed is None or isinstance(seed, int):
        seed = np.random.default_rng(seed)
    ests = model.modelfit_results.individual_estimates
    covs = model.modelfit_results.individual_estimates_covariance
    if parameters is None:
        parameters = ests.columns
    ests = ests[parameters]
    samples = pd.DataFrame()
    for (idx, mu), sigma in zip(ests.iterrows(), covs):
        sigma = sigma[parameters].loc[parameters]
        sigma = nearest_posdef(sigma)
        id_samples = seed.multivariate_normal(mu.values, sigma.values, size=samples_per_id)
        id_df = pd.DataFrame(id_samples, columns=ests.columns)
        id_df.index = [idx] * len(id_df)  # ID as index
        samples = pd.concat((samples, id_df))
    return samples



import pharmpy.model
import pharmpy.results
import pharmpy.tools.common
import pharmpy.tools.modelfit


def fit(models):
    if isinstance(models, pharmpy.model.Model):
        models = [models]
        single = True
    else:
        single = False
    tool = pharmpy.tools.modelfit.Modelfit(models)
    tool.run()
    if single:
        return models[0]
    else:
        return models


def create_results(path, **kwargs):
    res = pharmpy.tools.common.create_results(path, **kwargs)
    return res


def read_results(path):
    res = pharmpy.results.read_results(path)
    return res



# The estimation steps in a model


class EstimationMethod:
    def __init__(self, method, cov=False):
        method = self._canonicalize_and_check_method(method)
        self._method = method
        self.cov = cov

    def _canonicalize_and_check_method(self, method):
        method = method.upper()
        supported = ['FO', 'FOI', 'FOCE', 'FOCEI']
        if method not in supported:
            raise ValueError(f'Estimation method: {method} not recognized. Use any of {supported}.')
        return method

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        method = self._canonicalize_and_check_method(value)
        self._method = method

    def __eq__(self, other):
        return self._method == other._method and self.cov == other.cov

    def __repr__(self):
        return f'EstimationMethod("{self._method}", cov={self.cov})'



from pyfakefs.fake_filesystem_unittest import Patcher

import pharmpy.tools.modelsearch as ms
from pharmpy import Model
from pharmpy.plugins.nonmem import conf


def test_modelsearch(datadir):
    assert conf  # Make linter happy. Don't know why conf needs to be imported
    with Patcher(additional_skip_names=['pkgutil']) as patcher:
        fs = patcher.fs
        fs.add_real_file(datadir / 'pheno_real.mod', target_path='run1.mod')
        fs.add_real_file(datadir / 'pheno_real.phi', target_path='run1.phi')
        fs.add_real_file(datadir / 'pheno_real.lst', target_path='run1.lst')
        fs.add_real_file(datadir / 'pheno_real.ext', target_path='run1.ext')
        fs.add_real_file(datadir / 'pheno.dta', target_path='pheno.dta')
        model = Model('run1.mod')

        tool = ms.ModelSearch(model, 'stepwise', 'ABSORPTION(FO)')
        assert tool



import pharmpy.execute as execute
import pharmpy.results
import pharmpy.search.algorithms as algorithms
import pharmpy.search.rankfuncs as rankfuncs
import pharmpy.tools
import pharmpy.tools.modelfit as modelfit


class ModelSearch(pharmpy.tools.Tool):
    def __init__(self, base_model, algorithm, mfl, rankfunc='ofv', **kwargs):
        self.base_model = base_model
        self.mfl = mfl
        self.algorithm = getattr(algorithms, algorithm)
        self.rankfunc = getattr(rankfuncs, rankfunc)
        super().__init__(**kwargs)

    def fit(self, models):
        db = execute.LocalDirectoryDatabase(self.rundir.path / 'models')
        modelfit_run = modelfit.Modelfit(models, database=db, path=self.rundir.path)
        modelfit_run.run()

    def run(self):
        df = self.algorithm(
            self.base_model,
            self.mfl,
            self.fit,
            self.rankfunc,
        )
        res = ModelSearchResults(runs=df)
        res.to_json(path=self.rundir.path / 'results.json')
        res.to_csv(path=self.rundir.path / 'results.csv')
        return res


class ModelSearchResults(pharmpy.results.Results):
    def __init__(self, runs=None):
        self.runs = runs


def run_modelsearch(base_model, algorithm, mfl, **kwargs):
    ms = ModelSearch(base_model, algorithm, mfl, **kwargs)
    res = ms.run()
    return res



"""Common modeling pipeline elements
:meta private:
"""

from io import StringIO

from pharmpy.model_factory import Model


def read_model(path):
    """Read model from file"""
    model = Model(path)
    return model


def read_model_from_string(code):
    """Read model directly from the model code in a string"""
    model = Model(StringIO(code))
    return model


def write_model(model, path='', force=False):
    """Write model to file"""
    model.write(path=path, force=force)
    return model


def convert_model(model, to):
    """Convert model to other format

    Parameters
    ----------
    model : Model
        Model to convert
    to : str
        Name of format to convert into. Currently supported 'nlmixr'

    Results
    -------
    Model
        New model object with new underlying model format.
    """
    if to != 'nlmixr':
        raise ValueError(f"Unknown format {to}: supported format is 'nlmixr'")
    import pharmpy.plugins.nlmixr.model as nlmixr

    new = nlmixr.convert_model(model)
    return new


def update_source(model):
    """Update source

    Let the code of the underlying source language be updated to reflect
    changes in the model object.
    """
    model.update_source()
    return model


def copy_model(model):
    """Copies model to a new model object"""
    return model.copy()


def set_name(model, name):
    """Sets name of model object"""
    model.name = name
    return model


def set_initial_estimates(model, inits):
    """Set initial estimates

    Parameters
    ----------
    model : Model
        Pharmpy model
    inits : dict
        A dictionary of parameter init for parameters to change

    Returns
    -------
    model : Model
    """
    model.parameters.inits = inits
    return model


def fix_parameters(model, parameter_names):
    """Fix parameters

    Fix all listed parameters

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_names : list or str
        one parameter name or a list of parameter names

    Returns
    -------
    model : Model
    """
    if isinstance(parameter_names, str):
        d = {parameter_names: True}
    else:
        d = {name: True for name in parameter_names}
    model.parameters.fix = d
    return model


def unfix_parameters(model, parameter_names):
    """Unfix parameters

    Unfix all listed parameters

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_names : list or str
        one parameter name or a list of parameter names

    Returns
    -------
    model : Model
    """
    if isinstance(parameter_names, str):
        d = {parameter_names: False}
    else:
        d = {name: False for name in parameter_names}
    model.parameters.fix = d
    return model


def fix_parameters_to(model, parameter_names, values):
    """Fix parameters to

    Fix all listed parameters to specified value/values

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_names : list or str
        one parameter name or a list of parameter names
    values : list or int
        one value or a list of values (must be equal to number of parameter_names)

    Returns
    -------
    model : Model
    """
    if not parameter_names:
        parameter_names = [p.name for p in model.parameters]

    fix_parameters(model, parameter_names)
    d = _create_init_dict(parameter_names, values)
    set_initial_estimates(model, d)

    return model


def unfix_parameters_to(model, parameter_names, values):
    """Unix parameters to

    Unfix all listed parameters to specified value/values

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_names : list or str
        one parameter name or a list of parameter names
    values : list or int
        one value or a list of values (must be equal to number of parameter_names)

    Returns
    -------
    model : Model
    """
    if not parameter_names:
        parameter_names = [p.name for p in model.parameters]

    unfix_parameters(model, parameter_names)
    d = _create_init_dict(parameter_names, values)
    set_initial_estimates(model, d)

    return model


def _create_init_dict(parameter_names, values):
    if isinstance(parameter_names, str):
        d = {parameter_names: values}
    else:
        if not isinstance(values, list):
            values = [values] * len(parameter_names)
        if len(parameter_names) != len(values):
            raise ValueError(
                'Incorrect number of values, must be equal to number of parameters '
                '(or one if same for all)'
            )
        d = {name: value for name, value in zip(parameter_names, values)}

    return d



"""
=============================
Generic Sources/Resources I/O
=============================

Source code/resource manager for a :class:`~pharmpy.generic.Model` class implementation.

Noteworthy
----------

Module is target for *all* generalizable and non-agnostic code that concerns:

    1. Reading *in* streams/file-like resources.
    2. Formatting/generating source code from current state of :class:`~pharmpy.generic.Model`.
    3. Writing *out* source code to streams/file-like resources.
"""

import io
from pathlib import Path


def Source(obj):
    """Factory to create source object"""
    if isinstance(obj, str) or isinstance(obj, Path) or isinstance(obj, io.IOBase):
        src = FileSource(obj)
        return src


class SourceBase:
    """API to manage the original source object of a model.

    Is a model API attached to attribute :attr:`Model.source <pharmpy.generic.Model.source>`.

    The source format can be a text file, but it could also be any kind of object, i.e. an R object.
    """

    def __init__(self, obj):
        self.obj = obj
        self.code = self.read(obj)

    def write(self, path, force=False):
        """Write source to file."""
        if not force and path.exists():
            raise FileExistsError(f'Cannot overwrite model at {path} with "force" not set')
        with open(path, 'w', encoding='latin-1') as fp:
            fp.write(self.code)
        self.path = path


class FileSource(SourceBase):
    """Source formats for files
    property: filename_extension    (includes the dot)
    """

    def __init__(self, obj):
        if isinstance(obj, str) or isinstance(obj, Path):
            path = Path(obj)
            self.filename_extension = path.suffix
        else:
            self.filename_extension = ''
        super().__init__(obj)

    def read(self, path_or_io):
        """Read source from io object or from str path or path object"""
        if isinstance(path_or_io, str) or isinstance(path_or_io, Path):
            self.path = Path(path_or_io)
            with open(path_or_io, 'r', encoding='latin-1') as fp:
                return fp.read()
        else:
            self.path = Path('.')
            return path_or_io.read()



# Find private modules to exclude and run sphinx-apidoc

import os
import re
from pathlib import Path

import importlib
import pkgutil
import inspect

import pharmpy

pharmpy_path = Path(pharmpy.__file__).parent

# Remove old rst files to make sure no old lingers
os.system('rm docs/reference/*.rst')

exclude = []
noclass = []
for _, modname, is_pkg in pkgutil.walk_packages([Path('src/pharmpy')], 'pharmpy.'):
    module = importlib.import_module(modname)
    if hasattr(module, '__doc__') and module.__doc__ is not None and re.search(':meta private:', module.__doc__):
        path = Path(module.__file__).relative_to(pharmpy_path)
        exclude.append('src/pharmpy/' + str(path))
    else:
        no_classes = True
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj):
                qualname = obj.__module__ + '.' + obj.__name__
                if qualname.startswith(module.__name__):
                    no_classes = False
                    break
        if no_classes or is_pkg:        # No classes in subpackages to get rid of warnings about inheritance diagrams
            noclass.append(module.__name__)

noclass.append('pharmpy')

cmd = f'sphinx-apidoc -e -f -d 2 -o docs/reference src/pharmpy {" ".join(exclude)}'

os.system(cmd)

for path in Path('docs/reference').glob('pharmpy*.rst'):
    documented_module = str(path.name)[0:-4]
    if documented_module not in noclass:
        with open(path, 'a') as fh:
            print(f'''
Inheritance Diagram
-------------------

.. inheritance-diagram:: {documented_module}
   :parts: 4
''', file=fh)



import pytest

from pharmpy.search.mfl import ModelFeatures


@pytest.mark.parametrize(
    'code,args',
    [
        ('ABSORPTION(FO)', {'FO'}),
        ('ABSORPTION(* )', {'FO', 'ZO', 'SEQ-ZO-FO'}),
        ('ABSORPTION([ZO,FO])', {'FO', 'ZO'}),
        ('ABSORPTION([ZO,  FO])', {'FO', 'ZO'}),
        ('ABSORPTION( [   SEQ-ZO-FO,  FO   ]  )', {'SEQ-ZO-FO', 'FO'}),
        ('absorption([zo, fo])', {'FO', 'ZO'}),
        ('ABSORPTION(FO);ABSORPTION(ZO)', {'FO', 'ZO'}),
        ('ABSORPTION(FO)\nABSORPTION([FO, SEQ-ZO-FO])', {'FO', 'SEQ-ZO-FO'}),
    ],
)
def test_mfl_absorption(code, args):
    mfl = ModelFeatures(code)
    assert mfl.absorption.args == args


@pytest.mark.parametrize(
    'code,args',
    [
        ('ELIMINATION(FO)', {'FO'}),
        ('ELIMINATION( *)', {'FO', 'ZO', 'MM', 'MIX-FO-MM'}),
        ('ELIMINATION([ZO,FO])', {'FO', 'ZO'}),
        ('ELIMINATION([ZO,  FO])', {'FO', 'ZO'}),
        ('ELIMINATION( [   MIX-FO-MM,  FO   ]  )', {'MIX-FO-MM', 'FO'}),
        ('elimination([zo, fo])', {'FO', 'ZO'}),
        ('ELIMINATION(FO);ABSORPTION(ZO)', {'FO'}),
    ],
)
def test_mfl_elimination(code, args):
    mfl = ModelFeatures(code)
    assert mfl.elimination.args == args


@pytest.mark.parametrize(
    'code,args',
    [
        ('TRANSITS(0)', {0}),
        ('TRANSITS([0, 1])', {0, 1}),
        ('TRANSITS([0, 2, 4])', {0, 2, 4}),
        ('TRANSITS(0..1)', {0, 1}),
        ('TRANSITS(1..4)', {1, 2, 3, 4}),
        ('TRANSITS(1..4); TRANSITS(5)', {1, 2, 3, 4, 5}),
        ('TRANSITS(0);PERIPHERALS(0)', {0}),
    ],
)
def test_mfl_transits(code, args):
    mfl = ModelFeatures(code)
    assert mfl.transits.args == args


@pytest.mark.parametrize(
    'code,args',
    [
        ('PERIPHERALS(0)', {0}),
        ('PERIPHERALS([0, 1])', {0, 1}),
        ('PERIPHERALS([0, 2, 4])', {0, 2, 4}),
        ('PERIPHERALS(0..1)', {0, 1}),
        ('PERIPHERALS(1..4)', {1, 2, 3, 4}),
        ('PERIPHERALS(1..4); PERIPHERALS(5)', {1, 2, 3, 4, 5}),
    ],
)
def test_mfl_peripherals(code, args):
    mfl = ModelFeatures(code)
    assert mfl.peripherals.args == args


@pytest.mark.parametrize(
    'code,args',
    [
        ('LAGTIME()', None),
    ],
)
def test_mfl_lagtime(code, args):
    mfl = ModelFeatures(code)
    assert mfl.lagtime.args == args


@pytest.mark.parametrize(
    'code',
    [
        ('ABSORPTION(ILLEGAL)'),
        ('ELIMINATION(ALSOILLEGAL)'),
        ('LAGTIME(0)'),
        ('TRANSITS(*)'),
    ],
)
def test_illegal_mfl(code):
    with pytest.raises(Exception):
        ModelFeatures(code)



# import dateutil.parser
import re

from numpy import nan
from packaging import version


class NONMEMResultsFile:
    """Representing and parsing a NONMEM results file (aka lst-file)
    This is not a generic output object and will be combined by other classes
    into new structures.
    We rely on tags in NONMEM >= 7.2.0
    """

    def __init__(self, path=None):
        self.table = dict()
        self.nonmem_version = None
        if path is not None:
            for name, content in NONMEMResultsFile.table_blocks(path):
                if name == 'INIT':
                    self.nonmem_version = content.pop('nonmem_version', None)
                else:
                    self.table[name] = content

    @property
    def _supported_nonmem_version(self):
        return NONMEMResultsFile.supported_version(self.nonmem_version)

    def estimation_status(self, table_number):
        result = NONMEMResultsFile.unknown_termination()
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                for key in result.keys():
                    result[key] = self.table[table_number].get(key)
            else:
                result['minimization_successful'] = False
        return result

    def covariance_status(self, table_number):
        result = NONMEMResultsFile.unknown_covariance()
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                for key in result.keys():
                    result[key] = self.table[table_number].get(key)
            else:
                result['covariance_step_ok'] = False
        return result

    def ofv(self, table_number):
        ofv = None
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                ofv = self.table[table_number].get('OBJV')
        if ofv is not None:
            ofv = float(ofv)
        else:
            ofv = nan
        return ofv

    @staticmethod
    def supported_version(nonmem_version):
        return nonmem_version is not None and version.parse(nonmem_version) >= version.parse(
            '7.2.0'
        )

    @staticmethod
    def unknown_covariance():
        return {'covariance_step_ok': None}

    @staticmethod
    def unknown_termination():
        return {
            'minimization_successful': None,
            'estimate_near_boundary': None,
            'rounding_errors': None,
            'maxevals_exceeded': None,
            'significant_digits': nan,
            'function_evaluations': nan,
            'warning': None,
        }

    @staticmethod
    def cleanup_version(v):
        if v == 'V':
            v = '5.0'
        elif v == 'VI':
            v = '6.0'
        return v

    @staticmethod
    def parse_covariance(rows):
        result = NONMEMResultsFile.unknown_covariance()
        result['covariance_step_ok'] = False
        if len(rows) < 1:
            return result

        not_ok = re.compile(r' INTERPRET VARIANCE-COVARIANCE OF ESTIMATES WITH CARE')
        not_ok_either = re.compile(r'(R|S) MATRIX ALGORITHMICALLY SINGULAR')
        ok = re.compile(r' Elapsed covariance\s+time in seconds: ')  # need variable whitespace

        for row in rows:
            if not_ok.match(row):
                result['covariance_step_ok'] = False
                break
            if not_ok_either.search(row):
                result['covariance_step_ok'] = False
                break
            if ok.match(row):
                result['covariance_step_ok'] = True
                break
        return result

    @staticmethod
    def parse_termination(rows):
        result = NONMEMResultsFile.unknown_termination()
        if len(rows) < 1:  # will happen if e.g. TERMINATED BY OBJ during estimation
            result['minimization_successful'] = False
            return result
        result['estimate_near_boundary'] = False
        result['rounding_errors'] = False
        result['maxevals_exceeded'] = False
        result['warning'] = False

        success = [
            re.compile(r'0MINIMIZATION SUCCESSFUL'),
            re.compile(
                r' (REDUCED STOCHASTIC PORTION|OPTIMIZATION|'
                + r'BURN-IN|EXPECTATION ONLY PROCESS)'
                + r'( STATISTICALLY | WAS | )(COMPLETED|NOT TESTED)'
            ),
            re.compile(r'1OBJECTIVE FUNCTION IS TO BE EVALUATED'),
        ]
        failure = [
            re.compile(r'0MINIMIZATION TERMINATED'),
            re.compile(r'0SEARCH WITH ESTIMATION STEP WILL NOT PROCEED'),
        ]
        maybe = re.compile(
            r' (REDUCED STOCHASTIC PORTION|OPTIMIZATION|'
            + r'BURN-IN|EXPECTATION ONLY PROCESS)'
            + r'( WAS | )NOT COMPLETED'
        )  # success only if next line USER INTERRUPT
        misc = {
            'estimate_near_boundary': re.compile(
                r'0(ESTIMATE OF THETA IS NEAR THE BOUNDARY AND|'
                + r'PARAMETER ESTIMATE IS NEAR ITS BOUNDARY)'
            ),
            'rounding_errors': re.compile(r' DUE TO ROUNDING ERRORS'),
            'maxevals_exceeded': re.compile(r' DUE TO MAX. NO. OF FUNCTION EVALUATIONS EXCEEDED'),
            'warning': re.compile(r' HOWEVER, PROBLEMS OCCURRED WITH THE MINIMIZATION.'),
        }
        sig_digits = re.compile(r' NO. OF SIG. DIGITS IN FINAL EST.:\s*(\S+)')  # only classical est
        feval = re.compile(r' NO. OF FUNCTION EVALUATIONS USED:\s*(\S+)')  # only classical est

        maybe_success = False
        for row in rows:
            if maybe_success:
                result['minimization_successful'] = bool(re.search(r'USER INTERRUPT', row))
                break
            for p in success:
                if p.match(row):
                    result['minimization_successful'] = True
                    break
            if result['minimization_successful'] is not None:
                break
            for p in failure:
                if p.match(row):
                    result['minimization_successful'] = False
                    break
            if result['minimization_successful'] is not None:
                break
            maybe_success = bool(maybe.match(row))
        for row in rows:
            if sig_digits.match(row):
                result['significant_digits'] = float(sig_digits.match(row).group(1))
                continue
            if feval.match(row):
                result['function_evaluations'] = int(feval.match(row).group(1))
                continue
            for name, p in misc.items():
                if p.match(row):
                    result[name] = True
                    break
        return result

    @staticmethod
    def tag_items(path):
        nmversion = re.compile(r'1NONLINEAR MIXED EFFECTS MODEL PROGRAM \(NONMEM\) VERSION\s+(\S+)')
        tag = re.compile(r'\s*#([A-Z]{4}):\s*(.*)')
        end_TERE = re.compile(r'(0|1)')  # The part we need after #TERE: will preceed next ^1 or ^0
        cleanup = re.compile(r'\*+\s*')
        TERM = list()
        TERE = list()
        found_TERM = False
        found_TERE = False

        with open(path) as file:
            version_number = None
            for row in file:
                m = nmversion.match(row)
                if m:
                    version_number = NONMEMResultsFile.cleanup_version(m.group(1))
                    yield ('nonmem_version', version_number)
                    break  # we will stay at current file position
            if NONMEMResultsFile.supported_version(version_number):
                for row in file:
                    m = tag.match(row)
                    if m:
                        if m.group(1) == 'TERM':
                            if found_TERM:
                                raise NotImplementedError('Two TERM tags without TERE in between')
                            found_TERM = True
                            TERM = list()
                        elif m.group(1) == 'TERE':
                            if not found_TERM:
                                raise NotImplementedError('TERE tag without TERM tag')
                            found_TERE = True
                            yield ('TERM', NONMEMResultsFile.parse_termination(TERM))
                            found_TERM = False
                            TERM = list()
                        elif found_TERE:
                            raise NotImplementedError('TERE tag without ^1 or ^0 before next tag')
                        else:
                            v = cleanup.sub('', m.group(2))
                            yield (m.group(1), v.strip())
                    elif found_TERE:
                        if end_TERE.match(row):
                            yield ('TERE', NONMEMResultsFile.parse_covariance(TERE))
                            found_TERE = False
                            TERE = list()
                        else:
                            TERE.append(row)
                    elif found_TERM:
                        TERM.append(row)
                if found_TERM:
                    yield ('TERM', NONMEMResultsFile.parse_termination(TERM))
                if found_TERE:
                    yield ('TERE', NONMEMResultsFile.parse_covariance(TERE))

    @staticmethod
    def table_blocks(path):
        block = dict()
        table_number = 'INIT'
        for name, content in NONMEMResultsFile.tag_items(path):
            if name == 'TERM' or name == 'TERE':
                for k, v in content.items():
                    block[k] = v
            elif name == 'TBLN':
                if bool(block):
                    yield (table_number, block)
                block = dict()
                table_number = int(content)
            else:
                # If already set then it means TBLN was missing, probably $SIM, skip
                if name not in block.keys():
                    block[name] = content
        if bool(block):
            yield (table_number, block)



class Workflow:
    def __init__(self, tasks=None, models=None):
        self.tasks = []
        if tasks:
            self.add_tasks(tasks)
        self.models = models
        self.infiles = []
        self.outfiles = []

    def add_tasks(self, tasks):
        if isinstance(tasks, list):
            self.tasks.extend(tasks)
        else:
            self.tasks.append(tasks)

    def as_dict(self):
        as_dict = dict()
        for task in self.tasks:
            if isinstance(task.task_input, tuple):
                value = (task.function, *task.task_input)
            else:
                value = (task.function, task.task_input)
            as_dict[task.name] = value
        return as_dict

    def get_task(self, name):
        for task in self.tasks:
            if task.name == name:
                return task
        return None

    def add_infiles(self, source, destination='.'):
        self.infiles.append((source, destination))

    def add_outfiles(self, outfiles):
        self.outfiles.append(outfiles)


class Task:
    def __init__(self, name, function, task_input, condition=None):
        self.name = name
        self.function = function
        self.task_input = task_input
        self.condition = condition

    def add_condition(self, other):
        if self.condition and not isinstance(self.condition, list):
            self.condition = list(self.condition)
        self.condition.append(other)

    def __str__(self):
        return f'{self.name}: {self.function}({self.task_input})'



"""

.. list-table:: Options for the results module
   :widths: 25 25 50 150
   :header-rows: 1

   * - Option name
     - Default value
     - Type
     - Description
   * - ``native_shrinkage``
     - ``True``
     - bool
     - Should shrinkage calculation of external tool be used.
       Otherwise pharmpy will calculate shrinkage
"""

import json
import lzma
import math
import warnings
from pathlib import Path

import altair as alt
import numpy as np
import pandas as pd
import sympy

import pharmpy.config as config
import pharmpy.visualization
from pharmpy.data import PharmDataFrame
from pharmpy.data_structures import OrderedSet
from pharmpy.math import cov2corr
from pharmpy.parameter_sampling import sample_from_covariance_matrix


class ResultsConfiguration(config.Configuration):
    module = 'pharmpy.results'
    """ FIXME: This setting should be moved to the NONMEM plugin
    """
    native_shrinkage = config.ConfigItem(True, 'Use shrinkage results from external tool')


conf = ResultsConfiguration()


class ResultsJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pd.DataFrame) or isinstance(obj, pd.Series):
            d = json.loads(obj.to_json(orient='table'))
            if isinstance(obj, pd.DataFrame):
                d['__class__'] = 'DataFrame'
            elif isinstance(obj, pd.Series):
                d['__class__'] = 'Series'
            else:
                d['__class__'] = obj.__class__.__name__
            return d
        elif obj.__class__.__module__.startswith('altair.'):
            d = obj.to_dict()
            d['__class__'] = 'vega-lite'
            return d
        else:
            return json.JSONEncoder.encode(self, obj)


class ResultsJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
        self.cls = None

    def object_hook(self, dct):
        if '__class__' in dct:
            cls = dct['__class__']
            del dct['__class__']
            if cls == 'DataFrame' or cls == 'Series':
                res = pd.read_json(json.dumps(dct), orient='table')
                return res
            elif cls == 'vega-lite':
                res = alt.Chart.from_dict(dct)
                return res
            else:
                self.cls = cls
        return dct


def read_results(path_or_buf):
    if '{' in str(path_or_buf):  # Heuristic to determine if path or buffer
        s = path_or_buf
    else:
        path = Path(path_or_buf)
        if path.is_dir():
            path /= 'results.json'
        if not path.is_file():
            raise FileNotFoundError(str(path))
        if path.name.endswith('.xz'):
            with lzma.open(path, 'r') as json_file:
                s = json_file.read().decode('utf-8')
        else:
            with open(path, 'r') as json_file:
                s = json_file.read()
    decoder = ResultsJSONDecoder()
    d = decoder.decode(s)
    if decoder.cls == 'FREMResults':
        from pharmpy.tools.frem import FREMResults

        res = FREMResults.from_dict(d)
    elif decoder.cls == 'BootstrapResults':
        from pharmpy.tools.bootstrap import BootstrapResults

        res = BootstrapResults.from_dict(d)
    elif decoder.cls == 'CDDResults':
        from pharmpy.tools.cdd import CDDResults

        res = CDDResults.from_dict(d)
    elif decoder.cls == 'SCMResults':
        from pharmpy.tools.scm import SCMResults

        res = SCMResults.from_dict(d)
    elif decoder.cls == 'QAResults':
        from pharmpy.tools.qa import QAResults

        res = QAResults.from_dict(d)
    elif decoder.cls == 'LinearizeResults':
        from pharmpy.tools.linearize import LinearizeResults

        res = LinearizeResults.from_dict(d)
    return res


class Results:
    """Base class for all result classes"""

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    def to_json(self, path=None, lzma=False):
        json_dict = self.to_dict()
        json_dict['__class__'] = self.__class__.__name__
        s = json.dumps(json_dict, cls=ResultsJSONEncoder)
        if path:
            if not lzma:
                with open(path, 'w') as fh:
                    fh.write(s)
            else:
                xz_path = path.parent / (path.name + '.xz')
                with lzma.open(xz_path, 'w') as fh:
                    fh.write(bytes(s, 'utf-8'))
        else:
            return s

    def get_and_reset_index(self, attr, **kwargs):
        """Wrapper to reset index of attribute or result from method.

        Used to facilitate importing multiindex dataframes into R
        """
        val = getattr(self, attr)
        if callable(val):
            df = val(**kwargs)
        else:
            df = val
        return df.reset_index()

    def to_dict(self):
        """Convert results object to a dictionary"""
        return vars(self).copy()

    def __str__(self):
        start = self.__class__.__name__
        s = f'{start}\n\n'
        d = self.to_dict()
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                s += value.to_string()
            elif isinstance(value, list):  # Print list of lists as table
                if len(value) > 0 and isinstance(value[0], list):
                    df = pd.DataFrame(value)
                    df_str = df.to_string(index=False)
                    df_str = df_str.split('\n')[1:]
                    s += '\n'.join(df_str)
            else:
                s += str(value) + '\n'
            s += '\n\n'
        return s

    def to_csv(self, path):
        """Save results as a human readable csv file

        Index will not be printed if it is a basic range.
        """
        d = self.to_dict()
        s = ""
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                if isinstance(value.index, pd.RangeIndex):
                    use_index = False
                else:
                    use_index = True
                s += value.to_csv(index=use_index)
            elif isinstance(value, pd.Series):
                s += value.to_csv()
            elif isinstance(value, list):  # Print list of lists as table
                if len(value) > 0 and isinstance(value[0], list):
                    for row in value:
                        s += f'{",".join(map(str, row))}\n'
            else:
                s += str(value) + '\n'
            s += '\n'
        with open(path, 'w', newline='') as fh:
            print(s, file=fh)

    def create_report(self, path):
        import pharmpy.reporting.reporting as reporting

        reporting.generate_report(self.rst_path, path)

    def add_plots(self):
        """Create and add all plots to results object"""
        raise NotImplementedError()


class ModelfitResults(Results):
    """Base class for results from a modelfit operation

    properties: individual_OFV is a df with currently ID and iOFV columns
        model_name - name of model that generated the results
        model
    """

    def __init__(
        self,
        ofv=None,
        parameter_estimates=None,
        covariance_matrix=None,
        standard_errors=None,
        minimization_successful=None,
        individual_ofv=None,
        individual_estimates=None,
    ):
        self._ofv = ofv
        self._parameter_estimates = parameter_estimates
        self._covariance_matrix = covariance_matrix
        self._standard_errors = standard_errors
        self._minimization_successful = minimization_successful
        self._individual_estimates = individual_estimates
        self._individual_ofv = individual_ofv

    def to_dict(self):
        """Convert results object to a dictionary"""
        return {'parameter_estimates': self.parameter_estimates}

    def reparameterize(self, parameterizations):
        """Reparametrize all parameters given a list of parametrization object
        will change the parameter_estimates and standard_errors to be for
        the transformed parameter
        """
        raise NotImplementedError("Not implemented")

    @property
    def ofv(self):
        """Final objective function value"""
        return self._ofv

    @property
    def aic(self):
        """Final AIC value assuming the OFV to be -2LL"""
        parameters = self.model.parameters.copy()
        parameters.remove_fixed()
        return self.ofv + 2 * len(parameters)

    @property
    def bic(self):
        """Final BIC value assuming the OFV to be -2LL"""
        parameters = self.model.parameters.copy()
        parameters.remove_fixed()
        return self.ofv + len(parameters) * math.log(len(self.model.dataset.pharmpy.observations))

    @property
    def evaluation_ofv(self):
        """The ofv as if the model was evaulated

        Currently works for classical estimation methods by taking the OFV of the
        first iteration.
        """
        return self._evaluation_ofv

    @property
    def minimization_successful(self):
        """Was the minimization successful"""
        return self._minimization_successful

    @property
    def parameter_estimates(self):
        """Parameter estimates as series"""
        return self._parameter_estimates

    @parameter_estimates.setter
    def parameter_estimates(self, value):
        self._parameter_estimates = value

    @property
    def covariance_matrix(self):
        """The covariance matrix of the population parameter estimates"""
        return self._covariance_matrix

    def _cov_from_inf(self):
        Im = self.information_matrix
        cov = pd.DataFrame(np.linalg.inv(Im.values), index=Im.index, columns=Im.columns)
        return cov

    def _cov_from_corrse(self):
        se = self.standard_errors
        corr = self.correlation_matrix
        x = se.values @ se.values.T
        cov = x * corr.values
        cov_df = pd.DataFrame(cov, index=corr.index, columns=corr.columns)
        return cov_df

    @property
    def information_matrix(self):
        """The Fischer information matrix of the population parameter estimates"""
        raise NotImplementedError()

    def _inf_from_cov(self):
        C = self.covariance_matrix
        Im = pd.DataFrame(np.linalg.inv(C.values), index=C.index, columns=C.columns)
        return Im

    def _inf_from_corrse(self):
        se = self.standard_errors
        corr = self.correlation_matrix
        x = se.values @ se.values.T
        cov = x * corr.values
        Im = pd.DataFrame(np.linalg.inv(cov), index=corr.index, columns=corr.columns)
        return Im

    @property
    def correlation_matrix(self):
        """The correlation matrix of the population parameter estimates"""
        raise NotImplementedError()

    def _corr_from_cov(self):
        C = self.covariance_matrix
        corr = pd.DataFrame(cov2corr(C.values), index=C.index, columns=C.columns)
        return corr

    def _corr_from_inf(self):
        Im = self.information_matrix
        corr = pd.DataFrame(cov2corr(np.linalg.inv(Im.values)), index=Im.index, columns=Im.columns)
        return corr

    @property
    def standard_errors(self):
        """Standard errors of population parameter estimates"""
        return self._standard_errors

    @property
    def relative_standard_errors(self):
        """Relative standard errors of population parameter estimates"""
        if self.standard_errors is not None:
            ser = self.standard_errors / self.parameter_estimates
            ser.name = 'RSE'
            return ser

    def _se_from_cov(self):
        """Calculate the standard errors from the covariance matrix
        can be used by subclasses
        """
        cov = self.covariance_matrix
        se = pd.Series(np.sqrt(np.diag(cov.values)), index=cov.index)
        return se

    def _se_from_inf(self):
        """Calculate the standard errors from the information matrix"""
        Im = self.information_matrix
        se = pd.Series(np.sqrt(np.linalg.inv(Im.values)), index=Im.index)
        return se

    @property
    def individual_ofv(self):
        """A Series with individual estimates indexed over ID"""
        return self._individual_ofv

    @property
    def individual_estimates(self):
        """Individual parameter estimates

        A DataFrame with ID as index one column for each individual parameter
        """
        return self._individual_estimates

    @property
    def individual_estimates_covariance(self):
        """The covariance matrix of the individual estimates"""
        raise NotImplementedError("Not implemented")

    def eta_shrinkage(self, sd=False):
        """Eta shrinkage for each eta

        Variance = False to get sd scale
        """
        pe = self.parameter_estimates
        # Want parameter estimates combined with fixed parameter values
        param_inits = self.model.parameters.to_dataframe()['value']
        pe = pe.combine_first(param_inits)

        ie = self.individual_estimates
        param_names = self.model.random_variables.iiv.variance_parameters
        diag_ests = pe[param_names]
        diag_ests.index = ie.columns
        if not sd:
            shrinkage = 1 - (ie.var() / diag_ests)
        else:
            shrinkage = 1 - (ie.std() / (diag_ests ** 0.5))
        return shrinkage

    @property
    def individual_shrinkage(self):
        """The individual eta-shrinkage

        Definition: ieta_shr = (var(eta) / omega)
        """
        cov = self.individual_estimates_covariance
        pe = self.parameter_estimates
        # Want parameter estimates combined with fixed parameter values
        param_inits = self.model.parameters.to_dataframe()['value']
        pe = pe.combine_first(param_inits)

        # Get all iiv variance parameters
        param_names = self.model.random_variables.etas.variance_parameters
        param_names = list(OrderedSet(param_names))  # Only unique in order

        diag_ests = pe[param_names]

        def fn(row, ests):
            names = row[0].index
            ser = pd.Series(np.diag(row[0].values) / ests, index=names)
            return ser

        ish = pd.DataFrame(cov).apply(fn, axis=1, ests=diag_ests.values)
        return ish

    def near_bounds(self, zero_limit=0.001, significant_digits=2):
        return self.model.parameters.is_close_to_bound(
            values=self.parameter_estimates,
            zero_limit=zero_limit,
            significant_digits=significant_digits,
        )

    def parameter_summary(self):
        """Summary of parameter estimates and uncertainty"""
        pe = self.parameter_estimates
        ses = self.standard_errors
        rses = self.relative_standard_errors
        df = pd.DataFrame({'estimate': pe, 'SE': ses, 'RSE': rses})
        return df

    def plot_iofv_vs_iofv(self, other):
        x_label = f'{self.model_name} iOFV'
        y_label = f'{other.model_name} iOFV'
        df = PharmDataFrame({x_label: self.individual_ofv, y_label: other.individual_ofv})
        plot = pharmpy.visualization.scatter_plot_correlation(
            df, x_label, y_label, title='iOFV vs iOFV'
        )
        return plot

    def plot_individual_predictions(self, predictions=None, individuals=None):
        """Plot DV and predictions grouped on individuals

        Parameters
        ----------
        predictions : list
            A list of names of predictions to plot. None for all available
        individuals: list
            A list of individuals to include. None for all individuals
        """
        pred = self.predictions
        obs = self.model.dataset.pharmpy.observations
        indexcols = pred.index.names
        idcol = indexcols[0]
        idvcol = indexcols[1]

        data = pred.join(obs).reset_index()
        data = data.melt(id_vars=indexcols)

        if individuals is not None:
            data = data[data[idcol].isin(individuals)]
        if predictions is not None:
            dvcol = obs.name
            data = data[data['variable'].isin(predictions + [dvcol])]

        plot = (
            alt.Chart(data)
            .mark_line(point=True)
            .encode(x=idvcol, y='value', color='variable')
            .facet(f'{idcol}:N', columns=5)
        )
        return plot

    def individual_parameter_statistics(self, exprs, seed=None):
        """Calculate statistics for individual parameters

        exprs - is one string or an iterable of strings

        The parameter does not have to be in the model, but can be an
        expression of other parameters from the model.
        Does not support parameters that relies on the solution of the ODE-system
        """
        if isinstance(exprs, str) or isinstance(exprs, sympy.Basic):
            exprs = [_split_equation(exprs)]
        else:
            exprs = [_split_equation(expr) for expr in exprs]
        model = self.model
        dataset = model.dataset
        cols = set(dataset.columns)
        i = 0
        table = pd.DataFrame(columns=['parameter', 'covariates', 'mean', 'variance', 'stderr'])
        for name, expr in exprs:
            full_expr = model.statements.full_expression_from_odes(expr)
            covariates = {symb.name for symb in full_expr.free_symbols if symb.name in cols}
            if not covariates:
                cases = {'median': dict()}
            else:
                q5 = dataset[{'ID'} | covariates].groupby('ID').median().quantile(0.05)
                q95 = dataset[{'ID'} | covariates].groupby('ID').median().quantile(0.95)
                median = dataset[{'ID'} | covariates].groupby('ID').median().median()
                cases = {'p5': dict(q5), 'median': dict(median), 'p95': dict(q95)}

            df = pd.DataFrame(index=list(cases.keys()), columns=['mean', 'variance', 'stderr'])
            for case, cov_values in cases.items():
                pe = dict(model.modelfit_results.parameter_estimates)
                cov_expr = full_expr.subs(cov_values)
                expr = cov_expr.subs(pe)
                samples = model.random_variables.sample(
                    expr, parameters=pe, samples=1000000, seed=seed
                )

                mean = np.mean(samples)
                variance = np.var(samples)

                parameters = sample_from_covariance_matrix(
                    model,
                    n=100,
                    force_posdef_covmatrix=True,
                    seed=seed,
                )
                samples = []
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore')
                    for _, row in parameters.iterrows():
                        batch = model.random_variables.sample(
                            cov_expr.subs(dict(row)),
                            parameters=dict(row),
                            samples=10,
                            seed=seed,
                        )
                        samples.extend(list(batch))
                stderr = pd.Series(samples).std()
                df.loc[case] = [mean, variance, stderr]
                df.index.name = 'covariates'

            df.reset_index(inplace=True)
            if not name:
                name = f'unknown{i}'
                i += 1
            df['parameter'] = name
            table = pd.concat([table, df])
        table.set_index(['parameter', 'covariates'], inplace=True)
        return table

    def pk_parameters(self, seed=None):
        statements = self.model.statements
        odes = statements.ode_system
        central = odes.find_central()
        output = odes.find_output()
        depot = odes.find_depot(statements)
        peripherals = odes.find_peripherals()
        elimination_rate = odes.get_flow(central, output)

        expressions = []  # Eq(name, expr)
        # FO abs + 1comp + FO elimination
        if len(odes) == 3 and depot and odes.t not in elimination_rate.free_symbols:
            ode_list, ics = odes.to_explicit_odes(skip_output=True)
            sols = sympy.dsolve(ode_list, ics=ics)
            expr = sols[1].rhs
            d = sympy.diff(expr, odes.t)
            tmax_closed_form = sympy.solve(d, odes.t)[0]
            expressions.append(sympy.Eq(sympy.Symbol('t_max'), tmax_closed_form))
            e2 = sympy.simplify(expr / depot.dose.amount / sympy.denom(elimination_rate))
            cmax_dose_closed_form = sympy.simplify(e2.subs({odes.t: tmax_closed_form}))
            expressions.append(sympy.Eq(sympy.Symbol('C_max_dose'), cmax_dose_closed_form))

        # Any abs + 1comp + FO elimination
        if not peripherals and odes.t not in elimination_rate.free_symbols:
            elimination_system = statements.copy().ode_system
            # keep central and output
            for name in elimination_system.names:
                if name not in [central.name, output.name]:
                    elimination_system.remove_compartment(elimination_system.find_compartment(name))
                    ode_list, ics = elimination_system.to_explicit_odes(skip_output=True)
                    A0 = sympy.Symbol('A0')
                    ic = ics.popitem()
                    ics = {ic[0]: A0}
                    sols = sympy.dsolve(ode_list[0], ics=ics)
                    eq = sympy.Eq(sympy.Rational(1, 2) * A0, sols.rhs)
                    thalf_elim = sympy.solve(eq, odes.t)[0]
                    expressions.append(sympy.Eq(sympy.Symbol('t_half_elim'), thalf_elim))

        # Any abs + any comp + FO elimination
        if odes.t not in elimination_rate.free_symbols:
            expressions.append(sympy.Eq(sympy.Symbol('k_e'), elimination_rate))

        df = self.individual_parameter_statistics(expressions, seed=seed)
        return df


class ChainedModelfitResults(list, ModelfitResults):
    """A list of modelfit results given in order from first to final
    inherits from both list and ModelfitResults. Each method from ModelfitResults
    will be performed on the final modelfit object
    """

    def reparameterize(self, parameterizations):
        return self[-1].reparameterize(parameterizations)

    @property
    def ofv(self):
        return self[-1].ofv

    @property
    def evaluation_ofv(self):
        return self[0].evaluation_ofv

    @property
    def minimization_successful(self):
        return self[-1].minimization_successful

    @property
    def parameter_estimates(self):
        return self[-1].parameter_estimates

    @parameter_estimates.setter
    def parameter_estimates(self, value):
        self[-1].parameter_estimates = value

    @property
    def parameter_estimates_sdcorr(self):
        return self[-1].parameter_estimates_sdcorr

    @property
    def covariance_matrix(self):
        return self[-1].covariance_matrix

    @property
    def information_matrix(self):
        return self[-1].information_matrix

    @property
    def correlation_matrix(self):
        return self[-1].correlation_matrix

    @property
    def standard_errors(self):
        return self[-1].standard_errors

    @property
    def standard_errors_sdcorr(self):
        return self[-1].standard_errors_sdcorr

    @property
    def individual_ofv(self):
        return self[-1].individual_ofv

    @property
    def individual_estimates(self):
        return self[-1].individual_estimates

    @property
    def individual_estimates_covariance(self):
        return self[-1].individual_estimates_covariance

    @property
    def individual_shrinkage(self):
        return self[-1].individual_shrinkage

    @property
    def residuals(self):
        return self[-1].residuals

    @property
    def predictions(self):
        return self[-1].predictions

    def plot_iofv_vs_iofv(self, other):
        return self[-1].plot_iofv_vs_iofv(other)

    @property
    def model_name(self):
        return self[-1].model_name

    def individual_parameter_statistics(self, expr, seed=None):
        return self[-1].individual_parameter_statistics(expr, seed=seed)

    def pk_parameters(self, seed=None):
        return self[-1].pk_parameters(seed=seed)

    # FIXME: To not have to manually intercept everything here. Could do it in a general way.


def _split_equation(s):
    if isinstance(s, str):
        a = s.split('=')
        if len(a) == 1:
            name = None
            expr = sympy.sympify(s)
        else:
            name = a[0].strip()
            expr = sympy.sympify(a[1])
    elif isinstance(s, sympy.Eq):
        name = s.lhs.name
        expr = s.rhs
    else:  # sympy expr
        name = None
        expr = s
    if name is None and isinstance(expr, sympy.Symbol):
        name = expr.name
    return name, expr



import datetime

import pandas as pd


class Client:
    def __init__(self):
        self._log = []  # list of tuples of timestamp, category and message

    def log_error(self, message):
        self._log.append((datetime.datetime.now(), 'ERROR', message))

    def log_warning(self, message):
        self._log.append((datetime.datetime.now(), 'WARNING', message))

    def log_as_dataframe(self):
        times = [entry[0] for entry in self._log]
        categories = [entry[1] for entry in self._log]
        messages = [entry[2] for entry in self._log]
        df = pd.DataFrame({'time': times, 'category': categories, 'message': messages})
        return df



import pytest

from pharmpy.execute.databases.local_directory import LocalDirectoryDatabase
from pharmpy.execute.dispatcher import ExecutionDispatcher
from pharmpy.execute.dispatchers.local import LocalDispatcher
from pharmpy.tools.workflows import Task, Workflow


def fun(s):
    return s


def test_base_class():
    disp = ExecutionDispatcher()
    with pytest.raises(NotImplementedError):
        disp.run(None, None)


def test_local_dispatcher():
    db = LocalDirectoryDatabase()
    disp = LocalDispatcher()
    wfl = Task('results', fun, 'input')
    job = Workflow(wfl)
    job.models = []
    res = disp.run(job, db)
    assert res == 'input'



from pathlib import Path

import pharmpy.execute as execute

from .psn_helpers import tool_name


class RunDirectory:
    def __init__(self, method_name, path=None):
        i = 1
        while True:
            name = f'{method_name}_dir{i}'
            if path is not None:
                test_path = path / name
            else:
                test_path = Path(name)
            if not test_path.exists():
                test_path.mkdir()
                self.path = test_path
                self.models_path = test_path / 'models'
                self.models_path.mkdir()
                break
            i += 1


class Tool:
    def __init__(self, dispatcher=None, database=None, workflow_creator=None, path=None):
        self.rundir = RunDirectory(type(self).__name__.lower(), path=path)
        if dispatcher is None:
            self.dispatcher = execute.default_dispatcher
        else:
            self.dispatcher = dispatcher
        if database is None:
            self.database = execute.default_database
        else:
            self.database = database
        if workflow_creator is None:
            import pharmpy.plugins.nonmem.run

            self.workflow_creator = pharmpy.plugins.nonmem.run.create_workflow
        else:
            self.workflow_creator = workflow_creator


def create_results(path, **kwargs):
    name = tool_name(path)
    # FIXME: Do something automatic here
    if name == 'qa':
        from pharmpy.tools.qa.results import psn_qa_results

        res = psn_qa_results(path, **kwargs)
    elif name == 'bootstrap':
        from pharmpy.tools.bootstrap.results import psn_bootstrap_results

        res = psn_bootstrap_results(path, **kwargs)
    elif name == 'cdd':
        from pharmpy.tools.cdd.results import psn_cdd_results

        res = psn_cdd_results(path, **kwargs)
    elif name == 'frem':
        from pharmpy.tools.frem.results import psn_frem_results

        res = psn_frem_results(path, **kwargs)
    elif name == 'linearize':
        from pharmpy.tools.linearize.results import psn_linearize_results

        res = psn_linearize_results(path, **kwargs)
    elif name == 'scm':
        from pharmpy.tools.scm.results import psn_scm_results

        res = psn_scm_results(path, **kwargs)
    elif name == 'simeval':
        from pharmpy.tools.simeval.results import psn_simeval_results

        res = psn_simeval_results(path, **kwargs)
    else:
        raise ValueError("Not a valid run directory")
    return res



import pytest

from pharmpy.estimation import EstimationMethod


def test_estimation_method():
    a = EstimationMethod('foce', cov=True)
    assert a.method == 'FOCE'
    assert a.cov

    with pytest.raises(ValueError):
        EstimationMethod('sklarfs')

    a.method = 'fo'
    assert a.method == 'FO'

    assert a == EstimationMethod('fo', cov=True)
    assert repr(a) == 'EstimationMethod("FO", cov=True)'



import os
import subprocess
from pathlib import Path

from pharmpy.plugins.nonmem import conf
from pharmpy.tools.workflows import Task, Workflow


def create_workflow(models):
    wf = Workflow(models=models)
    task_names, execute_tasks = [], []

    for i, model in enumerate(models):
        execute_tasks.append(Task(f'run-{i}', execute_model, (model, i)))
        task_names.append(f'run-{i}')

    wf.add_tasks(execute_tasks)
    wf.add_tasks(Task('results', results, task_names))

    for i, model in enumerate(models):
        wf.add_infiles(model.dataset_path, destination=f'NONMEM_run{i}')
        wf.add_outfiles(
            [
                f'NONMEM_run{i}/{model.name}.lst',
                f'NONMEM_run{i}/{model.name}.ext',
                f'NONMEM_run{i}/{model.name}.phi',
            ]
        )
    return wf


def execute_model(model, i):
    path = Path(f'NONMEM_run{i}').resolve()
    model = model.copy()
    model.update_source()
    model.dataset_path = model.dataset_path.name  # Make path in $DATA local
    model.write(path=path, force=True)
    args = [
        nmfe_path(),
        model.name + model.source.filename_extension,
        str(Path(model.name).with_suffix('.lst')),
        f'-rundir={str(path)}',
    ]
    subprocess.call(
        args, stdin=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL
    )
    return model


def results(models):
    for model in models:
        model._modelfit_results = None
        model.modelfit_results.ofv
    return models


def nmfe_path():
    if os.name == 'nt':
        nmfe_candidates = ['nmfe74.bat', 'nmfe75.bat', 'nmfe73.bat']
    else:
        nmfe_candidates = ['nmfe74', 'nmfe75', 'nmfe73']
    path = conf.default_nonmem_path
    if path != Path(''):
        path /= 'run'
    for nmfe in nmfe_candidates:
        candidate_path = path / nmfe
        if candidate_path.is_file():
            path = candidate_path
            break
    else:
        raise FileNotFoundError(f'Cannot find nmfe script for NONMEM ({path})')
    return str(path)



"""
:meta private:
"""

from pharmpy.modeling.help_functions import _format_input_list, _get_etas
from pharmpy.symbols import symbol as S


def remove_iiv(model, list_to_remove=None):
    """
    Removes all IIV omegas given a list with eta names and/or parameter names.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    list_to_remove : str, list
        Name/names of etas and/or name/names of individual parameters to remove.
        If None, all etas that are IIVs will be removed. None is default.
    """
    rvs, sset = model.random_variables, model.statements
    list_to_remove = _format_input_list(list_to_remove)
    etas = _get_etas(model, list_to_remove, include_symbols=True)

    for eta in etas:
        statement = sset.find_assignment(eta.name, is_symbol=False)
        statement.expression = statement.expression.subs(S(eta.name), 0)
        del rvs[eta]

    model.random_variables = rvs
    model.statements = sset

    model.modelfit_results = None

    return model



import shutil

import pytest

import pharmpy.modeling as modeling
from pharmpy import Model
from pharmpy.config import site_config_dir, user_config_dir
from pharmpy.plugins.nonmem import conf
from pharmpy.utils import TemporaryDirectoryChanger


def test_configuration():
    print("User config dir:", user_config_dir())
    print("Site config dir:", site_config_dir())
    print("Default NONMEM path:", conf.default_nonmem_path)
    assert (conf.default_nonmem_path / 'license' / 'nonmem.lic').is_file()


def test_fit(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model = Model('pheno.mod')
        model.dataset_path = tmp_path / 'pheno.dta'
        modeling.fit(model)
        rundir = tmp_path / 'modelfit_dir1'
        assert model.modelfit_results.ofv == pytest.approx(730.8947268137308)
        assert rundir.is_dir()



import shutil
from pathlib import Path

from ..database import ModelDatabase


class LocalDirectoryDatabase(ModelDatabase):
    def __init__(self, path='.'):
        path = Path(path).resolve()
        if not path.exists():
            path.mkdir(parents=True)
        self.path = path

    def store_local_file(self, model, path):
        shutil.copy2(path, self.path)


class LocalModelDirectoryDatabase(ModelDatabase):
    def store_local_file(self, model, path):
        shutil.copy2(path, model.source.path.parent)
